#ifdef MASTERCAM_AFX
#include "stdafx.h"
#endif

#include "OperationMaker.h"

// Calls to this function can be eliminated from code by unsetting #MY_TRACE in OperationMaker.h
void operationmaker_function_trace(const char * fmt, ...)
{
#ifdef MY_TRACE
    if (pfLog == NULL)
	return;
    va_list ap;
    va_start(ap, fmt);

    char buff[10240];
    char * c = buff;
    vsprintf (c, fmt, ap);

    va_end(ap);

    static int count = 0;
//    printf ("Operation_maker\tCall#%d: %s\n", count, message);
    fprintf (pfLog, "Operation_maker\tCall#%d: %s\n", count, buff);
    count++;   
#endif
}

// Error reports can go to a message box, the trace file or nowhere depending on setting of #MY_REPORT and #MY_TRACE
void operationmaker_report_error(const char * fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);

    char buff[10240];
    char * c = buff;
    vsprintf (c, fmt, ap);

    va_end(ap);

	// Windows gadget to show a warning box requires windows.h
#ifdef MY_REPORT
    MessageBox (NULL, buff, "WARNING", MB_ICONSTOP);
#endif
#ifdef MY_TRACE
    if (pfLog == NULL)
	return;
    fprintf (pfLog, "Operation_maker\tERROR: %s\n", buff);
#endif
}

// Make a new project
Project * new_project (RoseDesign *des, const char *project_name)
{
    if (des == NULL || project_name == NULL) {
	REPORT_ERROR ("New project: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("New project name = %s in file %s", project_name, des->name());

    Project *project = Project::newInstance (des);
    project->put_its_id (project_name);
    return project;
}

// Make the main workplan of a project
// This function and the ones that follow make the data in the way shown in order to illustrate how the different types of objects must be threaded together
// You can also make a workplan and add it to a project later - the only real constraint is that both the project and plan must belong to the same STEP file
Workplan *main_workplan (Project *project, const char *plan_name) 
{					
    if (project == NULL || plan_name == NULL) {
	REPORT_ERROR ("Main workplan: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Workplan name = %s being added to project name = %s", plan_name, project->get_its_id());
 
    RoseDesign *des = project->getRootObject()->design();

    Workplan *plan = Workplan::newInstance (des);
    plan->put_its_id (plan_name);
    project->put_main_workplan (plan->getRoot());

    return plan;
}

// Make a workingstep and add it to the plan that will call that workingstep
Machining_workingstep *new_workingstep (Workplan *parent_plan, const char *ws_name) 
{
    if (parent_plan == NULL || ws_name == NULL) {
	REPORT_ERROR ("New workingstep: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Workingstep name = %s being added to workplan name = %s", ws_name, parent_plan->get_its_id());
    
    RoseDesign *des = parent_plan->getRootObject()->design();
    Machining_workingstep *ws = Machining_workingstep::newInstance (des);
    ws->put_its_id (ws_name);

    Manufacturing_feature_IF *default = find_default_feature_on_default_workpiece (des);
    ws->put_its_feature (ROSE_CAST (stp_shape_aspect, default->getRootObject()));
    parent_plan->add_its_elements (ws->getRoot());

    return ws;
}

// First look for a feature called default
// Second look for a workpiece called default and add a default feature
// Three create a default workpiece and add the default feature
Manufacturing_feature_IF *find_default_feature_on_default_workpiece (RoseDesign *des)
{
    ARMCursor cur;
    ARMObject * aobj;

    Manufacturing_feature_IF *result;

    // Try to find already made default feature
    cur.traverse(des);
    while ((aobj = cur.next())) {
	result = aobj->castToManufacturing_feature_IF();
	if (result != NULL && !strcmp ("default", result->get_its_id ()))
	    return result;
    }

    // Try to put new default feature on default workpiece
    Workpiece *piece = NULL;
    while ((aobj = cur.next())) {
	piece = aobj->castToWorkpiece();
	if (piece != NULL && !strcmp ("default", piece->get_its_id ())) {
	    result = new_toolpath_feature (piece, "default", new_simple_placement (0, 0, 0));
	    return result;
	}
    }

    // Make a default workpiece and add the default feature
    piece = Workpiece::newInstance (des);
    piece->put_its_id ("default");
    result = new_toolpath_feature (piece, "default", new_simple_placement (0, 0, 0));
    return result;
}



// A feed speed combination is represented as a technology
// Wherever sensible these should be shared between operations, between toolpaths and between both to enable easy editing
// There is an unused option to also set feedrate per tooth
Milling_technology *new_technology (RoseDesign *des, double feed, StixUnit feed_unit, double speed, StixUnit speed_unit)
{
     if (des == NULL || feed_unit == NULL || speed_unit == NULL) {
	REPORT_ERROR ("New technology: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Milling technology feed = %g (%s) speed = %g (%s)", feed, stix_get_unit_name (feed_unit), speed, stix_get_unit_name (speed_unit));
     
    Milling_technology *tech = Milling_technology::newInstance (des);

    tech->put_spindle (stix_make_measure_item (des, speed, speed_unit));
    tech->put_feedrate (stix_make_measure_item (des, feed, feed_unit));
    tech->put_synchronize_spindle_with_feed ("false");
    tech->put_inhibit_feedrate_override ("false");
    tech->put_inhibit_spindle_override ("false");
    tech->put_feedrate_reference ("tcp");			// feedrate is with respect to tool center point, other option is cutter contact point (ccp)

    return tech;
}

// Try to find a technology with these parameters and if not found then make
Milling_technology *find_or_make_technology (RoseDesign *des, double feed, StixUnit feed_unit, double speed, StixUnit speed_unit)
{
     if (des == NULL || feed_unit == NULL || speed_unit == NULL) {
	REPORT_ERROR ("Find or make technology: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Find or make technology feed = %g (%s) speed = %g (%s)", feed, stix_get_unit_name (feed_unit), speed, stix_get_unit_name (speed_unit));

    Milling_technology *tech;
    ARMObject *obj;
    ARMCursor cur;
    cur.traverse (des);

    while ((obj = cur.next()) != NULL) {
	if (obj->castToMilling_technology ()) {
	    tech = obj->castToMilling_technology ();
	    double tech_feed = stix_get_measure_value (tech->get_feedrate ());
	    double tech_speed = stix_get_measure_value (tech->get_spindle ());
	    StixUnit tech_feed_unit = stix_get_unit_type (tech->get_feedrate ());
	    StixUnit tech_speed_unit = stix_get_unit_type (tech->get_spindle ());
	    if (tech_feed == feed && tech_speed == speed && tech_feed_unit == feed_unit && tech_speed_unit == speed_unit)
		return tech;
	}
    }

    return new_technology (des, feed, feed_unit, speed, speed_unit);
}


// Functions can be set for the operation or the toolpath or both
// These are the required properties see the standard for documentation of the other properties
Milling_machine_functions *new_functions (RoseDesign *des, const char *coolant, const char *mist, const char*through_spindle, const char *chip_removal)
{
    if (des == NULL) {
	REPORT_ERROR ("New functions: function called with NULL arguments");
	return NULL;
    }    
    TRACE_CALL ("Milling machine functions coolant = %s, mist = %s, through_spindle = %s, chip_removal = %s", coolant, mist, through_spindle, chip_removal);
    
    Milling_machine_functions *funs = Milling_machine_functions::newInstance (des);

    funs->put_coolant (coolant);
    funs->put_mist (mist);
    funs->put_through_spindle_coolant (through_spindle);
    funs->put_chip_removal (chip_removal);

    return funs;
}

Milling_machine_functions *find_or_make_functions (RoseDesign *des, const char *coolant, const char *mist, const char*through_spindle, const char *chip_removal)
{
    if (des == NULL || coolant == NULL || mist == NULL || through_spindle == NULL || chip_removal == NULL) {
	REPORT_ERROR ("New functions: function called with NULL arguments");
	return NULL;
    }    
    TRACE_CALL ("Find or make milling machine functions coolant = %s, mist = %s, through_spindle = %s, chip_removal = %s", coolant, mist, through_spindle, chip_removal);
    
    Milling_machine_functions *funs;
    ARMObject *obj;
    ARMCursor cur;
    cur.traverse (des);

    while ((obj = cur.next()) != NULL) {
	if (obj->castToMilling_machine_functions ()) {
	    funs = obj->castToMilling_machine_functions ();
	    const char *funs_coolant = funs->get_coolant ();
	    const char *funs_mist = funs->get_mist ();
	    const char *funs_through = funs->get_through_spindle_coolant ();
	    const char *funs_chip = funs->get_chip_removal ();

	    if (!strcmp (funs_coolant, coolant) && !strcmp (funs_mist, mist) && !strcmp (funs_through, through_spindle) && !strcmp (funs_chip, chip_removal))
		return funs;
	}
    }

    return new_functions (des, coolant, mist, through_spindle, chip_removal);
}

Machining_operation_IF *new_freeform_operation (Machining_workingstep *ws, Milling_machine_cutting_tool_IF *tool, Milling_technology *tech, Milling_machine_functions *funs) 
{
     if (ws == NULL || tool == NULL || tech == NULL || funs == NULL) {
	REPORT_ERROR ("New freeform operation: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("New freeform operation for ws = %s feed = %g (%s) speed = %g (%s)", ws->get_its_id (), stix_get_measure_value (tech->get_feedrate ()), stix_get_unit_name (tech->get_feedrate ()),
			stix_get_measure_value (tech->get_spindle ()), stix_get_unit_name (tech->get_spindle ()));

    RoseDesign *des = ws->getRootObject()->design();

    Freeform_operation *op = Freeform_operation::newInstance (des);
    op->put_its_id (ws->get_its_id());						    // required but not usually shown
    op->put_its_tool (ROSE_CAST (stp_machining_tool, tool->getRootObject()));	    // Cast is annoying but necessary because tool is abstract
    op->put_its_machine_functions (funs->getRoot());
    op->put_its_technology (tech->getRoot ());

    ws->put_its_operation (op->getRoot());
    return op;
}

void set_toolpath_units (Machining_operation_IF *op, StixUnit length_unit, StixUnit angle_unit)
{
     if (op == NULL || length_unit == NULL || angle_unit == NULL) {
	REPORT_ERROR ("Set toolpath unit: function called with NULL arguments");
	return;
    }
    RoseDesign *des = op->getRootObject()->design();
    set_geometric_context (des, stix_make_geometry_context (des, "", 3, length_unit, angle_unit, stixunit_steradian));
}


Cutter_location_trajectory *new_toolpath (Machining_operation_IF *op, const char *type, const char *name)
{
     if (op == NULL ||  type == NULL) {
	REPORT_ERROR ("New toolpath: function called with NULL arguments");
	return NULL;
    }

    RoseDesign *des = op->getRootObject()->design();
    if (get_geometric_context (des) == NULL) {
	REPORT_ERROR ("New toolpath: no units set for new toolpath (see set_toolpath_units)");
	return NULL;
    }
    TRACE_CALL ("New toolpath for operation %s unit = %s",  op->get_its_id (), stix_get_unit_name (stix_get_context_length_unit (get_geometric_context (des))));

    Cutter_location_trajectory * path = Cutter_location_trajectory::newInstance (des);
    path->put_its_id (name);
    path->put_its_priority (EN_SUGGESTED);					// Other option is EN_REQUIRED for when cannot be altered
    path->put_its_type (type);							// EN_APPROACH, EN_LIFT, EN_CONNECT, EN_CONTACT, EN_NON_CONTACT, EN_TRAJECTORY_PATH (default)

    op->add_its_toolpath (path->getRoot ());
    return path;
}

Cutter_location_trajectory *new_toolpath_rapid (Machining_operation_IF *op, const char* type, const char *name)
{
    if (op == NULL || type == NULL) {
	REPORT_ERROR ("New toolpath rapid: function called with NULL arguments");
	return NULL;
    }
 
    RoseDesign *des = op->getRootObject()->design();
    if (get_geometric_context (des) == NULL) {
	REPORT_ERROR ("New toolpath rapid: no units set for new toolpath (see set_toolpath_units)");
	return NULL;
    }
    TRACE_CALL ("New rapid toolpath for operation %s unit = %s",  op->get_its_id (), stix_get_unit_name (stix_get_context_length_unit (get_geometric_context (des))));

    Cutter_location_trajectory * path = Cutter_location_trajectory::newInstance (des);
    path->put_its_id (name);
    path->put_its_priority (EN_SUGGESTED);					// Other option is EN_REQUIRED for when cannot be altered
    path->put_its_type (type);							// EN_APPROACH, EN_LIFT, EN_CONNECT, EN_CONTACT, EN_NON_CONTACT, EN_TRAJECTORY_PATH (default)
    path->set_rapid_speed ();

    op->add_its_toolpath (path->getRoot ());
    return path;
}

// Set functions for a toolpath when different from those of the operation
Toolpath_IF *set_toolpath_functions (Toolpath_IF *path, Milling_machine_functions *funs)
{
    if (path == NULL || funs == NULL) {
	REPORT_ERROR ("Set toolpath functions: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Setting explicit toolpath functions for path %s",  path->get_its_id ());

    path->put_its_machine_functions (funs->getRoot ());
    return path;
}

// Set technology for a toolpath when different from those of the operation
Toolpath_IF *set_toolpath_technology (Toolpath_IF *path, Milling_technology *tech)
{
    if (path == NULL || tech == NULL) {
	REPORT_ERROR ("Set toolpath functions: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Setting explicit toolpath technology for path %s to %g (%s) feedrate and %g (%s) spindle speed",  path->get_its_id (), stix_get_measure_value (tech->get_feedrate ()), stix_get_unit_name (tech->get_feedrate ()),
			stix_get_measure_value (tech->get_spindle ()), stix_get_unit_name (tech->get_spindle ()));

    path->put_its_technology (tech->getRoot ());
    return path;
}

Toolpath_IF *set_toolpath_maximum_deviation (Toolpath_IF *path, double value, StixUnit length_unit)
{
    if (path == NULL  || length_unit == NULL) {
	REPORT_ERROR ("Set toolpath functions: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Setting maximum deviation for path %s to %g (%s)",  path->get_its_id (), value, stix_get_unit_name (length_unit));

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    stp_measure_representation_item *max = stix_make_measure_item (des, value, length_unit);
    if (center) {
	center->put_path_maximum_deviation (max);
    }
    else if (contact) {
	contact->put_path_maximum_deviation (max);
    }
    else {
	REPORT_ERROR ("Set path maximum deviation: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    return path;
}

// Priority for toolpath (can it be deleted (EN_SUGGESTED the OperationMaker default) or is it required (EN_REQUIRED))
Toolpath_IF *set_toolpath_priority (Toolpath_IF *path, const char * priority)
{
    if (path == NULL || priority == NULL) {
	REPORT_ERROR ("Set toolpath priority: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("New priority for toolpath %s is %s",  path->get_its_id (), priority);

    path->put_its_priority (priority);
    return path;
}

stp_cartesian_point *start_tool_tip_curve (Toolpath_IF *path, double x, double y, double z)
{
     if (path == NULL) {
	REPORT_ERROR ("Start tool tip curve: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Starting tool tip curve for path %s at (%g, %g, %g)",  path->get_its_id (), x, y, z);

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_basiccurve ();
    else if (contact)
	obj = contact->get_basiccurve ();
    else {
	REPORT_ERROR ("Start tool tip curve: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj) {
	REPORT_ERROR ("Start tool tip curve: path has already been started");
	return NULL;
    }

    stp_polyline *curve = pnewIn(des) stp_polyline;
    curve->name("");
    stp_cartesian_point *first = pnewIn (des) stp_cartesian_point;
    first->name ("");
    first->coordinates()->add (x)->add (y)->add (z);
    curve->points()->add (first);

    if (center)
	center->put_basiccurve (curve);
    else if (contact)
	contact->put_basiccurve (curve);

    return first;
}

stp_cartesian_point *start_tool_tip_curve (Toolpath_IF *path, stp_cartesian_point *first)
{
     if (path == NULL || first == NULL) {
	REPORT_ERROR ("Start tool tip curve: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Starting tool tip curve for path %s at (repeat)",  path->get_its_id ());

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_basiccurve ();
    else if (contact)
	obj = contact->get_basiccurve ();
    else {
	REPORT_ERROR ("Start tool tip curve: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj) {
	REPORT_ERROR ("Start tool tip curve: path has already been started");
	return NULL;
    }

    stp_polyline *curve = pnewIn(des) stp_polyline;
    curve->name("tip");
    curve->points()->add (first);

    if (center)
	center->put_basiccurve (curve);
    else if (contact)
	contact->put_basiccurve (curve);

    return first;
}

stp_cartesian_point * goto_xyz (Toolpath_IF *path, double x, double y, double z)
{
     if (path == NULL) {
	REPORT_ERROR ("Goto xyz: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("goto_xyz (%g, %g, %g)",  x, y, z);

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_basiccurve ();
    else if (contact)
	obj = contact->get_basiccurve ();
    else {
	REPORT_ERROR ("Goto xyz: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj == NULL) {
	REPORT_ERROR ("Goto xyz: path has not been started");
	return NULL;
    }

    if (!obj->isa (ROSE_DOMAIN (stp_polyline))) {
	REPORT_ERROR ("Goto xyz: not processing composite curves yet");
	return NULL;
    }

    stp_polyline *curve = ROSE_CAST (stp_polyline, obj);
    stp_cartesian_point *next = pnewIn (des) stp_cartesian_point;
    next->name ("");
    next->coordinates()->add (x)->add (y)->add (z);
    curve->points()->add (next);

    return next;
}

stp_cartesian_point * goto_xyz (Toolpath_IF *path, stp_cartesian_point *next)
{
     if (path == NULL || next == NULL) {
	REPORT_ERROR ("Goto xyz: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("goto_xyz (repeat)");

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_basiccurve ();
    else if (contact)
	obj = contact->get_basiccurve ();
    else {
	REPORT_ERROR ("Goto xyz: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj == NULL) {
	REPORT_ERROR ("Goto xyz: path has not been started");
	return NULL;
    }

    if (!obj->isa (ROSE_DOMAIN (stp_polyline))) {
	REPORT_ERROR ("Goto xyz: not processing composite curves yet");
	return NULL;
    }

    stp_polyline *curve = ROSE_CAST (stp_polyline, obj);
    curve->points()->add (next);

    return next;
}


// via point to create an arc from pt[-1] and pt[+1]
// if next point is a via then this is a full circle
// via cannot be last point in polygon
stp_via_arc_point *goto_arc (Toolpath_IF *path, double x, double y, double z)			    // via point to create an arc from pt[-1] and pt[+1]
{
     if (path == NULL) {
	REPORT_ERROR ("Goto xyz via arc: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Goto_xyz_via_arc (%g, %g, %g)",  x, y, z);

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_basiccurve ();
    else if (contact)
	obj = contact->get_basiccurve ();
    else {
	REPORT_ERROR ("Goto xyz via arc: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj == NULL) {
	REPORT_ERROR ("Goto xyz via arc: path has not been started");
	return NULL;
    }

    if (!obj->isa (ROSE_DOMAIN (stp_polyline))) {
	REPORT_ERROR ("Goto xyz via arc: not processing composite curves yet");
	return NULL;
    }

    stp_polyline *curve = ROSE_CAST (stp_polyline, obj);
    stp_via_arc_point *next = pnewIn (des) stp_via_arc_point;
    next->name ("via_arc");
    next->coordinates()->add (x)->add (y)->add (z);
    curve->points()->add (next);

    return next;
}

stp_via_helix_point *goto_helix (Toolpath_IF *path, double x, double y, double z)			    // like an arc only depth changes wrt axis (not necessarily z)
{
     if (path == NULL) {
	REPORT_ERROR ("Goto xyz via helix: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Goto_xyz_via_helix (%g, %g, %g)",  x, y, z);

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_basiccurve ();
    else if (contact)
	obj = contact->get_basiccurve ();
    else {
	REPORT_ERROR ("Goto xyz via arc: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj == NULL) {
	REPORT_ERROR ("Goto xyz via arc: path has not been started");
	return NULL;
    }

    if (!obj->isa (ROSE_DOMAIN (stp_polyline))) {
	REPORT_ERROR ("Goto xyz via arc: not processing composite curves yet");
	return NULL;
    }

    stp_polyline *curve = ROSE_CAST (stp_polyline, obj);
    stp_via_helix_point *next = pnewIn (des) stp_via_helix_point;
    next->name ("via_helix");
    next->coordinates()->add (x)->add (y)->add (z);
    curve->points()->add (next);

    return next;
}

stp_cartesian_point *start_tool_axis_curve (Toolpath_IF *path, double i, double j, double k)
{
     if (path == NULL) {
	REPORT_ERROR ("Start tool axis curve: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Starting tool axis curve for path %s starts at (%g, %g, %g)",  path->get_its_id (), i, j, k);

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_its_toolaxis ();
    else if (contact)
	obj = contact->get_its_toolaxis ();
    else {
	REPORT_ERROR ("Start tool axis curve: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj) {
	REPORT_ERROR ("Start tool axis curve: path has already been started");
	return NULL;
    }

    stp_polyline *curve = pnewIn(des) stp_polyline;
    curve->name("axis");
    stp_cartesian_point *first = pnewIn (des) stp_cartesian_point;
    first->name ("");
    first->coordinates()->add (i)->add (j)->add (k);
    curve->points()->add (first);

    if (center)
	center->put_its_toolaxis (curve);
    else if (contact)
	contact->put_its_toolaxis (curve);

    return first;
}

stp_cartesian_point *start_tool_axis_curve (Toolpath_IF *path, stp_cartesian_point *first)
{
     if (path == NULL || first == NULL) {
	REPORT_ERROR ("Start tool axis curve: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Starting tool axis curve for path %s starts at (repeat axis)",  path->get_its_id ());

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_its_toolaxis ();
    else if (contact)
	obj = contact->get_its_toolaxis ();
    else {
	REPORT_ERROR ("Start tool axis curve: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj) {
	REPORT_ERROR ("Start tool axis curve: path has already been started");
	return NULL;
    }

    stp_polyline *curve = pnewIn(des) stp_polyline;
    curve->name("axis");
    curve->points()->add (first);

    if (center)
	center->put_its_toolaxis (curve);
    else if (contact)
	contact->put_its_toolaxis (curve);

    return first;
}

stp_cartesian_point * goto_axis (Toolpath_IF *path, double i, double j, double k)
{
     if (path == NULL) {
	REPORT_ERROR ("Goto axis: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("goto_axis (%g, %g, %g)",  i, j, k);

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_its_toolaxis ();
    else if (contact)
	obj = contact->get_its_toolaxis ();
    else {
	REPORT_ERROR ("Goto axis: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj == NULL) {
	REPORT_ERROR ("Goto axis: path has not been started");
	return NULL;
    }

    if (!obj->isa (ROSE_DOMAIN (stp_polyline))) {
	REPORT_ERROR ("Goto axis: not processing composite curves yet");
	return NULL;
    }

    stp_polyline *curve = ROSE_CAST (stp_polyline, obj);
    stp_cartesian_point *next = pnewIn (des) stp_cartesian_point;
    next->name ("");
    next->coordinates()->add (i)->add (j)->add (k);
    curve->points()->add (next);

    return next;
}

stp_cartesian_point * goto_axis (Toolpath_IF *path, stp_cartesian_point *next)
{
     if (path == NULL || next == NULL) {
	REPORT_ERROR ("Goto axis: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("goto_axis (repeat axis)");

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_its_toolaxis ();
    else if (contact)
	obj = contact->get_its_toolaxis ();
    else {
	REPORT_ERROR ("Goto axis: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj == NULL) {
	REPORT_ERROR ("Goto axis: path has not been started");
	return NULL;
    }

    if (!obj->isa (ROSE_DOMAIN (stp_polyline))) {
	REPORT_ERROR ("Goto axis: not processing composite curves yet");
	return NULL;
    }

    stp_polyline *curve = ROSE_CAST (stp_polyline, obj);
    curve->points()->add (next);

    return next;
}

stp_cartesian_point *start_surface_normal_curve (Toolpath_IF *path, double i, double j, double k)
{
     if (path == NULL) {
	REPORT_ERROR ("Start surface normal curve: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Starting surface normal curve for path %s starts at (%g, %g, %g)",  path->get_its_id (), i, j, k);

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_surface_normal ();
    else if (contact)
	obj = contact->get_surface_normal ();
    else {
	REPORT_ERROR ("Start normal curve: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj) {
	REPORT_ERROR ("Start surface normal curve: path has already been started");
	return NULL;
    }

    stp_polyline *curve = pnewIn(des) stp_polyline;
    curve->name("surface normal");
    stp_cartesian_point *first = pnewIn (des) stp_cartesian_point;
    first->name ("");
    first->coordinates()->add (i)->add (j)->add (k);
    curve->points()->add (first);

    if (center)
	center->put_surface_normal (curve);
    else if (contact)
	contact->put_surface_normal (curve);

    return first;
}

stp_cartesian_point *start_surface_normal_curve (Toolpath_IF *path, stp_cartesian_point *first)
{
     if (path == NULL || first == NULL) {
	REPORT_ERROR ("Start surface normal curve: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Starting surface normal curve for path %s starts at (repeat surface normal)",  path->get_its_id ());

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_surface_normal ();
    else if (contact)
	obj = contact->get_surface_normal ();
    else {
	REPORT_ERROR ("Start surface normal curve: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj) {
	REPORT_ERROR ("Start surface normal curve: path has already been started");
	return NULL;
    }

    stp_polyline *curve = pnewIn(des) stp_polyline;
    curve->name("surface normal");
    curve->points()->add (first);

    if (center)
	center->put_surface_normal (curve);
    else if (contact)
	contact->put_surface_normal (curve);

    return first;
}

stp_cartesian_point *goto_normal (Toolpath_IF *path, double i, double j, double k)
{
     if (path == NULL) {
	REPORT_ERROR ("Goto normal: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("goto_normal (%g, %g, %g)",  i, j, k);

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_surface_normal ();
    else if (contact)
	obj = contact->get_surface_normal ();
    else {
	REPORT_ERROR ("Goto normal: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj == NULL) {
	REPORT_ERROR ("Goto normal: path has not been started");
	return NULL;
    }

    if (!obj->isa (ROSE_DOMAIN (stp_polyline))) {
	REPORT_ERROR ("Goto normal: not processing composite curves yet");
	return NULL;
    }

    stp_polyline *curve = ROSE_CAST (stp_polyline, obj);
    stp_cartesian_point *next = pnewIn (des) stp_cartesian_point;
    next->name ("");
    next->coordinates()->add (i)->add (j)->add (k);
    curve->points()->add (next);

    return next;
}

stp_cartesian_point * goto_normal (Toolpath_IF *path, stp_cartesian_point *next)
{
     if (path == NULL || next == NULL) {
	REPORT_ERROR ("Goto normal: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("goto_normal (repeat surface normal)");

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_surface_normal ();
    else if (contact)
	obj = contact->get_surface_normal ();
    else {
	REPORT_ERROR ("Goto normal: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj == NULL) {
	REPORT_ERROR ("Goto normal: path has not been started");
	return NULL;
    }

    if (!obj->isa (ROSE_DOMAIN (stp_polyline))) {
	REPORT_ERROR ("Goto normal: not processing composite curves yet");
	return NULL;
    }

    stp_polyline *curve = ROSE_CAST (stp_polyline, obj);
    curve->points()->add (next);

    return next;
}

stp_cartesian_point *start_reference_direction_curve (Toolpath_IF *path, double i, double j, double k)
{
     if (path == NULL) {
	REPORT_ERROR ("Start reference direction curve: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Starting reference direction curve for path %s starts at (%g, %g, %g)",  path->get_its_id (), i, j, k);

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_its_toolref_direction ();
    else if (contact)
	obj = contact->get_its_toolref_direction ();
    else {
	REPORT_ERROR ("Start reference direction curve: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj) {
	REPORT_ERROR ("Start reference direction curve: path has already been started");
	return NULL;
    }

    stp_polyline *curve = pnewIn(des) stp_polyline;
    curve->name("reference direction");
    stp_cartesian_point *first = pnewIn (des) stp_cartesian_point;
    first->name ("");
    first->coordinates()->add (i)->add (j)->add (k);
    curve->points()->add (first);

    if (center)
	center->put_its_toolref_direction (curve);
    else if (contact)
	contact->put_its_toolref_direction (curve);

    return first;
}

stp_cartesian_point *start_reference_direction_curve (Toolpath_IF *path, stp_cartesian_point *first)
{
     if (path == NULL || first == NULL) {
	REPORT_ERROR ("Start reference direction curve: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Starting reference direction curve for path %s starts at (repeat reference direction)",  path->get_its_id ());

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_its_toolref_direction ();
    else if (contact)
	obj = contact->get_its_toolref_direction ();
    else {
	REPORT_ERROR ("Start reference direction curve: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj) {
	REPORT_ERROR ("Start reference direction curve: path has already been started");
	return NULL;
    }

    stp_polyline *curve = pnewIn(des) stp_polyline;
    curve->name("reference direction");
    curve->points()->add (first);

    if (center)
	center->put_its_toolref_direction (curve);
    else if (contact)
	contact->put_its_toolref_direction (curve);

    return first;
}

stp_cartesian_point *goto_direction (Toolpath_IF *path, double i, double j, double k)
{
     if (path == NULL) {
	REPORT_ERROR ("Goto direction: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("goto_direction (%g, %g, %g)",  i, j, k);

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_its_toolref_direction ();
    else if (contact)
	obj = contact->get_its_toolref_direction ();
    else {
	REPORT_ERROR ("Goto direction: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj == NULL) {
	REPORT_ERROR ("Goto direction: path has not been started");
	return NULL;
    }

    if (!obj->isa (ROSE_DOMAIN (stp_polyline))) {
	REPORT_ERROR ("Goto direction: not processing composite curves yet");
	return NULL;
    }

    stp_polyline *curve = ROSE_CAST (stp_polyline, obj);
    stp_cartesian_point *next = pnewIn (des) stp_cartesian_point;
    next->name ("");
    next->coordinates()->add (i)->add (j)->add (k);
    curve->points()->add (next);

    return next;
}

stp_cartesian_point * goto_direction (Toolpath_IF *path, stp_cartesian_point *next)
{
     if (path == NULL || next == NULL) {
	REPORT_ERROR ("Goto direction: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("goto_direction (repeat surface direction)");

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_its_toolref_direction ();
    else if (contact)
	obj = contact->get_its_toolref_direction ();
    else {
	REPORT_ERROR ("Goto direction: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj == NULL) {
	REPORT_ERROR ("Goto direction: path has not been started");
	return NULL;
    }

    if (!obj->isa (ROSE_DOMAIN (stp_polyline))) {
	REPORT_ERROR ("Goto direction: not processing composite curves yet");
	return NULL;
    }

    stp_polyline *curve = ROSE_CAST (stp_polyline, obj);
    curve->points()->add (next);

    return next;
}

stp_cartesian_point *start_cross_section_7d_curve (Toolpath_IF *path, double RC_max, double AC_max, double X_offset, double Y_offset, double RD_max, double RD_avg, double RD_offset)
{
     if (path == NULL) {
	REPORT_ERROR ("Start cross section curve: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Starting cross section curve for path %s starts at (%g, %g, %g, %g, %g, %g, %g)",  path->get_its_id (),  RC_max, AC_max, X_offset, Y_offset, RD_max, RD_avg, RD_offset);

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_cross_section_area_parameters ();
    else if (contact)
	obj = contact->get_cross_section_area_parameters ();
    else {
	REPORT_ERROR ("Start cross_section curve: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj) {
	REPORT_ERROR ("Start cross section curve: path has already been started");
	return NULL;
    }

    stp_polyline *curve = pnewIn(des) stp_polyline;
    curve->name("cross section");
    stp_cartesian_point *first = pnewIn (des) stp_cartesian_point;
    first->name ("");
    first->coordinates()->add (RC_max)->add (AC_max)->add (X_offset)->add (Y_offset)->add (RD_max)->add (RD_avg)->add (RD_offset);
    curve->points()->add (first);

    if (center)
	center->put_cross_section_area_parameters (curve);
    else if (contact)
	contact->put_cross_section_area_parameters (curve);

    return first;
}

stp_cartesian_point *start_cross_section_curve (Toolpath_IF *path, stp_cartesian_point *first)
{
     if (path == NULL || first == NULL) {
	REPORT_ERROR ("Start cross section curve: function called with NULL arguments");
	return NULL;
    }
    if (first->coordinates()->size() != 7) {
	REPORT_ERROR ("Start cross section curve: cross section points required to have 7 dimensions");
	return NULL;
    }
 
    TRACE_CALL ("Starting cross section curve for path %s starts at (repeat cross_section)",  path->get_its_id ());

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_cross_section_area_parameters ();
    else if (contact)
	obj = contact->get_cross_section_area_parameters ();
    else {
	REPORT_ERROR ("Start cross section curve: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj) {
	REPORT_ERROR ("Start cross section curve: path has already been started");
	return NULL;
    }

    stp_polyline *curve = pnewIn(des) stp_polyline;
    curve->name("cross section");
    curve->points()->add (first);

    if (center)
	center->put_cross_section_area_parameters (curve);
    else if (contact)
	contact->put_cross_section_area_parameters (curve);

    return first;
}

stp_cartesian_point *goto_cross_section (Toolpath_IF *path, double RC_max, double AC_max, double X_offset, double Y_offset, double RD_max, double RD_avg, double RD_offset)
{
     if (path == NULL) {
	REPORT_ERROR ("Goto cross_section: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("goto_cross_section (%g, %g, %g, %g, %g, %g, %g)", RC_max, AC_max, X_offset, Y_offset, RD_max, RD_avg, RD_offset);

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_cross_section_area_parameters ();
    else if (contact)
	obj = contact->get_cross_section_area_parameters ();
    else {
	REPORT_ERROR ("Goto cross_section: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj == NULL) {
	REPORT_ERROR ("Goto cross_section: path has not been started");
	return NULL;
    }

    if (!obj->isa (ROSE_DOMAIN (stp_polyline))) {
	REPORT_ERROR ("Goto cross_section: not processing composite curves yet");
	return NULL;
    }

    stp_polyline *curve = ROSE_CAST (stp_polyline, obj);
    stp_cartesian_point *next = pnewIn (des) stp_cartesian_point;
    next->name ("");
    next->coordinates()->add (RC_max)->add (AC_max)->add (X_offset)->add (Y_offset)->add (RD_max)->add (RD_avg)->add (RD_offset);
    curve->points()->add (next);

    return next;
}

stp_cartesian_point * goto_cross_section (Toolpath_IF *path, stp_cartesian_point *next)
{
     if (path == NULL || next == NULL) {
	REPORT_ERROR ("Goto cross_section: function called with NULL arguments");
	return NULL;
    }
    if (next->coordinates()->size() != 7) {
	REPORT_ERROR ("Start cross section curve: cross section points required to have 7 dimensions");
	return NULL;
    }    TRACE_CALL ("goto_cross_section (repeat cross_section)");

    RoseDesign *des = path->getRootObject()->design();
    RoseObject *obj = NULL;
    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    if (center)
	obj = center->get_cross_section_area_parameters ();
    else if (contact)
	obj = contact->get_cross_section_area_parameters ();
    else {
	REPORT_ERROR ("Goto cross_section: path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return NULL;
    }
    
    if (obj == NULL) {
	REPORT_ERROR ("Goto cross_section: path has not been started");
	return NULL;
    }

    if (!obj->isa (ROSE_DOMAIN (stp_polyline))) {
	REPORT_ERROR ("Goto cross_section: not processing composite curves yet");
	return NULL;
    }

    stp_polyline *curve = ROSE_CAST (stp_polyline, obj);
    curve->points()->add (next);

    return next;
}

Cutter_contact_trajectory *new_toolpath_contact (Machining_operation_IF *op, const char *type, const char *name)
{
     if (op == NULL || type == NULL) {
	REPORT_ERROR ("New toolpath contact: function called with NULL arguments");
	return NULL;
    }

    RoseDesign *des = op->getRootObject()->design();
    if (get_geometric_context (des) == NULL) {
	REPORT_ERROR ("New toolpath contact: no units set for new toolpath (see set_toolpath_units)");
	return NULL;
    }
    TRACE_CALL ("New contact toolpath for operation %s unit = %s",  op->get_its_id (), stix_get_unit_name (stix_get_context_length_unit (get_geometric_context (des))));

    Cutter_contact_trajectory * path = Cutter_contact_trajectory::newInstance (des);
    path->put_its_id (name);
    path->put_its_priority (EN_SUGGESTED);					// Other option is EN_REQUIRED for when cannot be altered
    path->put_its_type (type);							// EN_APPROACH, EN_LIFT, EN_CONNECT, EN_CONTACT, EN_NON_CONTACT, EN_TRAJECTORY_PATH (default)

    op->add_its_toolpath (path->getRoot ());
    return path;
}

Cutter_contact_trajectory *new_toolpath_contact_rapid (Machining_operation_IF *op, const char* type, const char *name)
{
    if (op == NULL || type == NULL) {
	REPORT_ERROR ("New toolpath contact rapid: function called with NULL arguments");
	return NULL;
    }

    RoseDesign *des = op->getRootObject()->design();
    if (get_geometric_context (des) == NULL) {
	REPORT_ERROR ("New toolpath contact rapid: no units set for new toolpath (see set_toolpath_units)");
	return NULL;
    }
    TRACE_CALL ("New contact rapid toolpath for operation %s unit = %s",  op->get_its_id (), stix_get_unit_name (stix_get_context_length_unit (get_geometric_context (des))));

    Cutter_contact_trajectory * path = Cutter_contact_trajectory::newInstance (des);
    path->put_its_id (name);
    path->put_its_priority (EN_SUGGESTED);					// Other option is EN_REQUIRED for when cannot be altered
    path->put_its_type (type);							// EN_APPROACH, EN_LIFT, EN_CONNECT, EN_CONTACT, EN_NON_CONTACT, EN_TRAJECTORY_PATH (default)
    path->set_rapid_speed ();

    op->add_its_toolpath (path->getRoot ());
    return path;
}

// check the curves in a toolpath for a consistent basis
// in current version this means same number of points
// but other types of curves are allowed in STEP
bool check_toolpath_curves (Toolpath_IF *path) 
{

    if (path == NULL) {
	REPORT_ERROR ("Check toolpath curves: function called with NULL arguments");
	return false;
    }

    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    RoseObject *obj = NULL;
    if (center) {
	obj = center->get_basiccurve ();
	TRACE_CALL ("Checking curves of toolpath %s for consistency", center->get_its_id());
    }
    else if (contact) {
	obj = contact->get_basiccurve ();
	TRACE_CALL ("Checking curves of toolpath %s for consistency", contact->get_its_id());
    }
    else {
	REPORT_ERROR ("Check toolpath curves : path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return false;
    }
    
    if (obj == NULL) {
	REPORT_ERROR ("Check toolpath curves: path has not been started");
	return false;
    }

    if (!obj->isa (ROSE_DOMAIN (stp_polyline))) {
	REPORT_ERROR ("Check toolpath curves: not processing composite curves yet");
	return false;
    }

    stp_polyline *curve = ROSE_CAST (stp_polyline, obj);
    if (curve->points()->size () < 2) {
	REPORT_ERROR ("Check toolpath curves: toolpath required to contain at least 2 points");
	return false;
    }

    if (center)
	obj = center->get_its_toolaxis ();
    else if (contact)
	obj = contact->get_its_toolaxis ();
    else
	obj = NULL;

    if (obj != NULL) {
	if (!obj->isa (ROSE_DOMAIN (stp_polyline))) {
	    REPORT_ERROR ("Check toolpath curves: toolaxis is not a polyline");
	    return false;
	}

	stp_polyline *second_curve = ROSE_CAST (stp_polyline, obj);
	if (second_curve->points()->size () != curve->points()->size()) {
	    REPORT_ERROR ("Check toolpath curves: toolpath curve and toolaxis curve have different basis");
	    return false;
	}
    }

    if (center)
	obj = center->get_its_toolref_direction ();
    else if (contact)
	obj = contact->get_its_toolref_direction ();
    else
	obj = NULL;

    if (obj != NULL) {
	if (!obj->isa (ROSE_DOMAIN (stp_polyline))) {
	    REPORT_ERROR ("Check toolpath curves: tool reference direction is not a polyline");
	    return false;
	}

	stp_polyline *second_curve = ROSE_CAST (stp_polyline, obj);
	if (second_curve->points()->size () != curve->points()->size()) {
	    REPORT_ERROR ("Check toolpath curves: toolpath curve and tool reference direction curve have different basis");
	    return false;
	}
    }

    if (center)
	obj = center->get_surface_normal ();
    else if (contact)
	obj = contact->get_surface_normal ();
    else
	obj = NULL;

    if (obj != NULL) {
	if (!obj->isa (ROSE_DOMAIN (stp_polyline))) {
	    REPORT_ERROR ("Check toolpath curves: surface normal is not a polyline");
	    return false;
	}

	stp_polyline *second_curve = ROSE_CAST (stp_polyline, obj);
	if (second_curve->points()->size () != curve->points()->size()) {
	    REPORT_ERROR ("Check toolpath curves: toolpath curve and surface normal curve have different basis");
	    return false;
	}
    }

    if (center)
	obj = center->get_cross_section_area_parameters ();
    else if (contact)
	obj = contact->get_cross_section_area_parameters ();
    else
	obj = NULL;

    if (obj != NULL) {
	if (!obj->isa (ROSE_DOMAIN (stp_polyline))) {
	    REPORT_ERROR ("Check toolpath curves: cross section is not a polyline");
	    return false;
	}

	stp_polyline *second_curve = ROSE_CAST (stp_polyline, obj);
	if (second_curve->points()->size () != curve->points()->size()) {
	    REPORT_ERROR ("Check toolpath curves: toolpath curve and cross section curve have different basis");
	    return false;
	}
    }

    return true;
}

// Specific type of machining operations with parameters for key values

Freeform_rough_milling *new_freeform_rough_milling (Machining_workingstep *ws, 
    Milling_machine_cutting_tool_IF *tool, Milling_technology *tech, Milling_machine_functions *funs) 
{
     if (ws == NULL || tool == NULL || tech == NULL || funs == NULL) {
	REPORT_ERROR ("New freeform rough milling operation: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("New freeform rough operation for ws = %s feed = %g (%s) speed = %g (%s)", ws->get_its_id (), stix_get_measure_value (tech->get_feedrate ()), stix_get_unit_name (tech->get_feedrate ()),
			stix_get_measure_value (tech->get_spindle ()), stix_get_unit_name (tech->get_spindle ()));

    RoseDesign *des = ws->getRootObject()->design();

    Freeform_rough_milling *op = Freeform_rough_milling::newInstance (des);
    op->put_its_id (ws->get_its_id());						    // required but not usually shown
    op->put_its_tool (ROSE_CAST (stp_machining_tool, tool->getRootObject()));	    // Cast is annoying but necessary because tool is abstract
    op->put_its_machine_functions (funs->getRoot());
    op->put_its_technology (tech->getRoot ());

    ws->put_its_operation (op->getRoot());
    return op;
}

Freeform_finish_milling *new_freeform_finish_milling (Machining_workingstep *ws, 
    Milling_machine_cutting_tool_IF *tool, Milling_technology *tech, Milling_machine_functions *funs)
{
     if (ws == NULL || tool == NULL || tech == NULL || funs == NULL) {
	REPORT_ERROR ("New freeform finish milling operation: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("New freeform finish operation for ws = %s feed = %g (%s) speed = %g (%s)", ws->get_its_id (), stix_get_measure_value (tech->get_feedrate ()), stix_get_unit_name (tech->get_feedrate ()),
			stix_get_measure_value (tech->get_spindle ()), stix_get_unit_name (tech->get_spindle ()));

    RoseDesign *des = ws->getRootObject()->design();

    Freeform_finish_milling *op = Freeform_finish_milling::newInstance (des);
    op->put_its_id (ws->get_its_id());						    // required but not usually shown
    op->put_its_tool (ROSE_CAST (stp_machining_tool, tool->getRootObject()));	    // Cast is annoying but necessary because tool is abstract
    op->put_its_machine_functions (funs->getRoot());
    op->put_its_technology (tech->getRoot ());

    ws->put_its_operation (op->getRoot());
    return op;
}   

Plane_rough_milling *new_plane_rough_milling (
	Machining_workingstep *ws, Milling_machine_cutting_tool_IF *tool, 
	 Milling_technology *tech, Milling_machine_functions *funs, 
	    double allowance, double axial_depth, StixUnit length_unit)
{
    if (ws == NULL || tool == NULL || tech == NULL || funs == NULL || length_unit == NULL) {
	REPORT_ERROR ("New plane rough milling operation: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("New plane rough milling operation for ws = %s feed = %g (%s) speed = %g (%s)", ws->get_its_id (), stix_get_measure_value (tech->get_feedrate ()), stix_get_unit_name (tech->get_feedrate ()),
			stix_get_measure_value (tech->get_spindle ()), stix_get_unit_name (tech->get_spindle ()));

    RoseDesign *des = ws->getRootObject()->design();

    Plane_rough_milling *op = Plane_rough_milling::newInstance (des);
    op->put_its_id (ws->get_its_id());						    // required but not usually shown
    op->put_its_tool (ROSE_CAST (stp_machining_tool, tool->getRootObject()));	    // Cast is annoying but necessary because tool is abstract
    op->put_its_machine_functions (funs->getRoot());
    op->put_its_technology (tech->getRoot ());
    op->put_allowance_bottom (stix_make_measure_item (des, allowance, length_unit));
    op->put_axial_cutting_depth (stix_make_measure_item (des, axial_depth, length_unit));

    ws->put_its_operation (op->getRoot());
    return op;
}

Plane_finish_milling *new_plane_finish_milling (
	    Machining_workingstep *ws, Milling_machine_cutting_tool_IF *tool, 
	    Milling_technology *tech, Milling_machine_functions *funs, 
	    double allowance, double axial_depth, StixUnit length_unit)
{
    if (ws == NULL || tool == NULL || tech == NULL || funs == NULL || length_unit == NULL) {
	REPORT_ERROR ("New plane finish milling operation: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("New plane finish milling operation for ws = %s feed = %g (%s) speed = %g (%s)", ws->get_its_id (), stix_get_measure_value (tech->get_feedrate ()), stix_get_unit_name (tech->get_feedrate ()),
			stix_get_measure_value (tech->get_spindle ()), stix_get_unit_name (tech->get_spindle ()));

    RoseDesign *des = ws->getRootObject()->design();

    Plane_finish_milling *op = Plane_finish_milling::newInstance (des);
    op->put_its_id (ws->get_its_id());						    // required but not usually shown
    op->put_its_tool (ROSE_CAST (stp_machining_tool, tool->getRootObject()));	    // Cast is annoying but necessary because tool is abstract
    op->put_its_machine_functions (funs->getRoot());
    op->put_its_technology (tech->getRoot ());
    op->put_allowance_bottom (stix_make_measure_item (des, allowance, length_unit));
    op->put_axial_cutting_depth (stix_make_measure_item (des, axial_depth, length_unit));

    ws->put_its_operation (op->getRoot());
    return op;
}

Bottom_and_side_rough_milling  *new_bottom_and_side_rough_milling (
	    Machining_workingstep *ws, Milling_machine_cutting_tool_IF *tool, 
	    Milling_technology *tech, Milling_machine_functions *funs, 
	    double side_allowance, double bottom_allowance, double radial_depth, 
	    double axial_depth, StixUnit length_unit)
{
    if (ws == NULL || tool == NULL || tech == NULL || funs == NULL || length_unit == NULL) {
	REPORT_ERROR ("New bottom and side rough milling operation: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("New bottom and side rough milling operation for ws = %s feed = %g (%s) speed = %g (%s)", ws->get_its_id (), stix_get_measure_value (tech->get_feedrate ()), stix_get_unit_name (tech->get_feedrate ()),
			stix_get_measure_value (tech->get_spindle ()), stix_get_unit_name (tech->get_spindle ()));

    RoseDesign *des = ws->getRootObject()->design();

    Bottom_and_side_rough_milling *op = Bottom_and_side_rough_milling::newInstance (des);
    op->put_its_id (ws->get_its_id());						    // required but not usually shown
    op->put_its_tool (ROSE_CAST (stp_machining_tool, tool->getRootObject()));	    // Cast is annoying but necessary because tool is abstract
    op->put_its_machine_functions (funs->getRoot());
    op->put_its_technology (tech->getRoot ());
    op->put_allowance_bottom (stix_make_measure_item (des, bottom_allowance, length_unit));
    op->put_allowance_side (stix_make_measure_item (des, side_allowance, length_unit));
    op->put_axial_cutting_depth (stix_make_measure_item (des, axial_depth, length_unit));
    op->put_radial_cutting_depth (stix_make_measure_item (des, radial_depth, length_unit));

    ws->put_its_operation (op->getRoot());
    return op;
}

Bottom_and_side_finish_milling *new_bottom_and_side_finish_milling (
	Machining_workingstep *ws, Milling_machine_cutting_tool_IF *tool, 
	Milling_technology *tech, Milling_machine_functions *funs, 
	double side_allowance, double bottom_allowance, double radial_depth, 
	double axial_depth, StixUnit length_unit)
{
    if (ws == NULL || tool == NULL || tech == NULL || funs == NULL || length_unit == NULL) {
	REPORT_ERROR ("New bottom and side finish milling operation: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("New bottom and side finish milling operation for ws = %s feed = %g (%s) speed = %g (%s)", ws->get_its_id (), stix_get_measure_value (tech->get_feedrate ()), stix_get_unit_name (tech->get_feedrate ()),
			stix_get_measure_value (tech->get_spindle ()), stix_get_unit_name (tech->get_spindle ()));

    RoseDesign *des = ws->getRootObject()->design();

    Bottom_and_side_finish_milling *op = Bottom_and_side_finish_milling::newInstance (des);
    op->put_its_id (ws->get_its_id());						    // required but not usually shown
    op->put_its_tool (ROSE_CAST (stp_machining_tool, tool->getRootObject()));	    // Cast is annoying but necessary because tool is abstract
    op->put_its_machine_functions (funs->getRoot());
    op->put_its_technology (tech->getRoot ());
    op->put_allowance_bottom (stix_make_measure_item (des, bottom_allowance, length_unit));
    op->put_allowance_side (stix_make_measure_item (des, side_allowance, length_unit));
    op->put_axial_cutting_depth (stix_make_measure_item (des, axial_depth, length_unit));
    op->put_radial_cutting_depth (stix_make_measure_item (des, radial_depth, length_unit));

    ws->put_its_operation (op->getRoot());
    return op;
}

Side_rough_milling *new_side_rough_milling (
	Machining_workingstep *ws, Milling_machine_cutting_tool_IF *tool, 
	Milling_technology *tech, Milling_machine_functions *funs, 
	double side_allowance, double radial_depth, double axial_depth, StixUnit length_unit)
{
    if (ws == NULL || tool == NULL || tech == NULL || funs == NULL || length_unit == NULL) {
	REPORT_ERROR ("New side rough milling operation: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("New side rough milling operation for ws = %s feed = %g (%s) speed = %g (%s)", ws->get_its_id (), stix_get_measure_value (tech->get_feedrate ()), stix_get_unit_name (tech->get_feedrate ()),
			stix_get_measure_value (tech->get_spindle ()), stix_get_unit_name (tech->get_spindle ()));

    RoseDesign *des = ws->getRootObject()->design();

    Side_rough_milling *op = Side_rough_milling::newInstance (des);
    op->put_its_id (ws->get_its_id());						    // required but not usually shown
    op->put_its_tool (ROSE_CAST (stp_machining_tool, tool->getRootObject()));	    // Cast is annoying but necessary because tool is abstract
    op->put_its_machine_functions (funs->getRoot());
    op->put_its_technology (tech->getRoot ());
    op->put_allowance_side (stix_make_measure_item (des, side_allowance, length_unit));
    op->put_axial_cutting_depth (stix_make_measure_item (des, axial_depth, length_unit));
    op->put_radial_cutting_depth (stix_make_measure_item (des, radial_depth, length_unit));

    ws->put_its_operation (op->getRoot());
    return op;
}

Side_finish_milling *new_side_finish_milling (
	Machining_workingstep *ws, Milling_machine_cutting_tool_IF *tool, 
	Milling_technology *tech, Milling_machine_functions *funs, 
	double side_allowance, double radial_depth, double axial_depth, StixUnit length_unit)
{
    if (ws == NULL || tool == NULL || tech == NULL || funs == NULL || length_unit == NULL) {
	REPORT_ERROR ("New side finish milling operation: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("New side finish milling operation for ws = %s feed = %g (%s) speed = %g (%s)", ws->get_its_id (), stix_get_measure_value (tech->get_feedrate ()), stix_get_unit_name (tech->get_feedrate ()),
			stix_get_measure_value (tech->get_spindle ()), stix_get_unit_name (tech->get_spindle ()));

    RoseDesign *des = ws->getRootObject()->design();

    Side_finish_milling *op = Side_finish_milling::newInstance (des);
    op->put_its_id (ws->get_its_id());						    // required but not usually shown
    op->put_its_tool (ROSE_CAST (stp_machining_tool, tool->getRootObject()));	    // Cast is annoying but necessary because tool is abstract
    op->put_its_machine_functions (funs->getRoot());
    op->put_its_technology (tech->getRoot ());
    op->put_allowance_side (stix_make_measure_item (des, side_allowance, length_unit));
    op->put_axial_cutting_depth (stix_make_measure_item (des, axial_depth, length_unit));
    op->put_radial_cutting_depth (stix_make_measure_item (des, radial_depth, length_unit));

    ws->put_its_operation (op->getRoot());
    return op;
}

// drilling operations
Drilling *new_drilling_operation (
	Machining_workingstep *ws, Milling_machine_cutting_tool_IF *tool, 
	Milling_technology *tech, Milling_machine_functions *funs, 
	double cutting_depth, double dwell_time_bottom, double feed_on_retract, 
	double previous_diameter, StixUnit length_unit, StixUnit time_unit)
{
    if (ws == NULL || tool == NULL || tech == NULL || funs == NULL || length_unit == NULL || time_unit == NULL) {
	REPORT_ERROR ("New drilling operation: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("New drilling operation for ws = %s feed = %g (%s) speed = %g (%s)", ws->get_its_id (), stix_get_measure_value (tech->get_feedrate ()), stix_get_unit_name (tech->get_feedrate ()),
			stix_get_measure_value (tech->get_spindle ()), stix_get_unit_name (tech->get_spindle ()));

    RoseDesign *des = ws->getRootObject()->design();

    Drilling *op = Drilling::newInstance (des);
    op->put_its_id (ws->get_its_id());						    // required but not usually shown
    op->put_its_tool (ROSE_CAST (stp_machining_tool, tool->getRootObject()));	    // Cast is annoying but necessary because tool is abstract
    op->put_its_machine_functions (funs->getRoot());
    op->put_its_technology (tech->getRoot ());
    if (cutting_depth > 0)
	op->put_cutting_depth (stix_make_measure_item (des, cutting_depth, length_unit));
    if (previous_diameter > 0)
	op->put_previous_diameter (stix_make_measure_item (des, previous_diameter, length_unit));
    if (dwell_time_bottom > 0)
	op->put_dwell_time_bottom (stix_make_measure_item (des, dwell_time_bottom, time_unit));
    if (feed_on_retract > 0)
	op->put_feed_on_retract (stix_make_ratio_measure_item (des, feed_on_retract, stixunit_ratio));

    ws->put_its_operation (op->getRoot());
    return op;
}

Multistep_drilling *new_multistep_drilling_operation (
	Machining_workingstep *ws, Milling_machine_cutting_tool_IF *tool, 
	Milling_technology *tech, Milling_machine_functions *funs, 
	double cutting_depth, double depth_of_step, double first_depth, double retract_distance,
	double dwell_time_bottom, double feed_on_retract, double dwell_time_step, double previous_diameter, 
	StixUnit length_unit, StixUnit time_unit)
{
        if (ws == NULL || tool == NULL || tech == NULL || funs == NULL || length_unit == NULL || time_unit == NULL) {
	REPORT_ERROR ("New multistep drilling operation: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("New multistep drilling operation for ws = %s feed = %g (%s) speed = %g (%s)", ws->get_its_id (), stix_get_measure_value (tech->get_feedrate ()), stix_get_unit_name (tech->get_feedrate ()),
			stix_get_measure_value (tech->get_spindle ()), stix_get_unit_name (tech->get_spindle ()));

    RoseDesign *des = ws->getRootObject()->design();

    Multistep_drilling *op = Multistep_drilling::newInstance (des);
    op->put_its_id (ws->get_its_id());						    // required but not usually shown
    op->put_its_tool (ROSE_CAST (stp_machining_tool, tool->getRootObject()));	    // Cast is annoying but necessary because tool is abstract
    op->put_its_machine_functions (funs->getRoot());
    op->put_its_technology (tech->getRoot ());
    if (cutting_depth > 0)
	op->put_cutting_depth (stix_make_measure_item (des, cutting_depth, length_unit));
    if (depth_of_step > 0)
	op->put_depth_of_step (stix_make_measure_item (des, depth_of_step, length_unit));
    if (first_depth > 0)
	op->put_first_depth (stix_make_measure_item (des, first_depth, length_unit));
    if (retract_distance > 0)
	op->put_retract_distance (stix_make_measure_item (des, retract_distance, length_unit));
    if (previous_diameter > 0)
	op->put_previous_diameter (stix_make_measure_item (des, previous_diameter, length_unit));
    if (dwell_time_bottom > 0)
	op->put_dwell_time_bottom (stix_make_measure_item (des, dwell_time_bottom, time_unit));
    if (dwell_time_step > 0)
	op->put_dwell_time_step (stix_make_measure_item (des, dwell_time_step, time_unit));
    if (feed_on_retract > 0)
	op->put_feed_on_retract (stix_make_ratio_measure_item (des, feed_on_retract, stixunit_ratio));

    ws->put_its_operation (op->getRoot());
    return op;
}

Tapping	*new_tapping_operation (
	Machining_workingstep *ws, Milling_machine_cutting_tool_IF *tool, 
	Milling_technology *tech, Milling_machine_functions *funs, 
	double cutting_depth, const char *compensation_chuck_required, 
	double dwell_time_bottom, double feed_on_retract, double previous_diameter, 
	StixUnit length_unit, StixUnit time_unit)
{
    if (ws == NULL || tool == NULL || tech == NULL || funs == NULL || length_unit == NULL || time_unit == NULL) {
	REPORT_ERROR ("New tapping operation: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("New tapping operation for ws = %s feed = %g (%s) speed = %g (%s)", ws->get_its_id (), stix_get_measure_value (tech->get_feedrate ()), stix_get_unit_name (tech->get_feedrate ()),
			stix_get_measure_value (tech->get_spindle ()), stix_get_unit_name (tech->get_spindle ()));

    RoseDesign *des = ws->getRootObject()->design();

    Tapping *op = Tapping::newInstance (des);
    op->put_its_id (ws->get_its_id());						    // required but not usually shown
    op->put_its_tool (ROSE_CAST (stp_machining_tool, tool->getRootObject()));	    // Cast is annoying but necessary because tool is abstract
    op->put_its_machine_functions (funs->getRoot());
    op->put_its_technology (tech->getRoot ());
    if (cutting_depth > 0)
	op->put_cutting_depth (stix_make_measure_item (des, cutting_depth, length_unit));
    if (previous_diameter > 0)
	op->put_previous_diameter (stix_make_measure_item (des, previous_diameter, length_unit));
    if (dwell_time_bottom > 0)
	op->put_dwell_time_bottom (stix_make_measure_item (des, dwell_time_bottom, time_unit));
    op->put_compensation_chuck (compensation_chuck_required);	    // EN_COMPENSATION_CHUCK_NOT_USED, EN_COMPENSATION_CHUCK_USED 
    if (feed_on_retract > 0)
	op->put_feed_on_retract (stix_make_ratio_measure_item (des, feed_on_retract, stixunit_ratio));

    ws->put_its_operation (op->getRoot());
    return op;
}

Reaming	*new_reaming_operation (
	Machining_workingstep *ws, Milling_machine_cutting_tool_IF *tool, Milling_technology *tech, Milling_machine_functions *funs, 
	double cutting_depth, double dwell_time_bottom, double feed_on_retract, double previous_diameter,
	const char *spindle_stop_at_bottom, double waiting_position_x, double waiting_position_y, double waiting_position_z, 
	double depth_of_testcut, StixUnit length_unit, StixUnit time_unit)
{
    if (ws == NULL || tool == NULL || tech == NULL || funs == NULL || length_unit == NULL || time_unit == NULL) {
	REPORT_ERROR ("New reaming operation: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("New reaming operation for ws = %s feed = %g (%s) speed = %g (%s)", ws->get_its_id (), stix_get_measure_value (tech->get_feedrate ()), stix_get_unit_name (tech->get_feedrate ()),
			stix_get_measure_value (tech->get_spindle ()), stix_get_unit_name (tech->get_spindle ()));

    RoseDesign *des = ws->getRootObject()->design();
    if (get_geometric_context (des) == NULL) {
	REPORT_ERROR ("New reaming operation: no units set for waiting point (see set_toolpath_units)");
	return NULL;
    }

    Reaming *op = Reaming::newInstance (des);
    op->put_its_id (ws->get_its_id());						    // required but not usually shown
    op->put_its_tool (ROSE_CAST (stp_machining_tool, tool->getRootObject()));	    // Cast is annoying but necessary because tool is abstract
    op->put_its_machine_functions (funs->getRoot());
    op->put_its_technology (tech->getRoot ());
    if (cutting_depth > 0)
	op->put_cutting_depth (stix_make_measure_item (des, cutting_depth, length_unit));
    if (depth_of_testcut > 0)
	op->put_depth_of_testcut (stix_make_measure_item (des, depth_of_testcut, length_unit));
    if (previous_diameter > 0)
	op->put_previous_diameter (stix_make_measure_item (des, previous_diameter, length_unit));
    if (dwell_time_bottom > 0)
	op->put_dwell_time_bottom (stix_make_measure_item (des, dwell_time_bottom, time_unit));
    op->put_spindle_stop_at_bottom (spindle_stop_at_bottom);	    // EN_SPINDLE_NONSTOP, EN_SPINDLE_STOP_AT_BOTTOM
    StixVctr waiting = StixVctr (waiting_position_x, waiting_position_y, waiting_position_z);
    op->put_waiting_position(waiting.makePtIn (des, ""));
    if (feed_on_retract > 0)
	op->put_feed_on_retract (stix_make_ratio_measure_item (des, feed_on_retract, stixunit_ratio));

    ws->put_its_operation (op->getRoot());
    return op;
}

Boring *new_boring_operation (
	Machining_workingstep *ws, Milling_machine_cutting_tool_IF *tool, 
	Milling_technology *tech, Milling_machine_functions *funs, 
	double cutting_depth, double dwell_time_bottom, double feed_on_retract, 
	double previous_diameter, const char *spindle_stop_at_bottom, 
	double waiting_position_x, double waiting_position_y, double waiting_position_z, 
	double depth_of_testcut, StixUnit length_unit, StixUnit time_unit)
{
    if (ws == NULL || tool == NULL || tech == NULL || funs == NULL || length_unit == NULL || time_unit == NULL) {
	REPORT_ERROR ("New boring operation: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("New boring operation for ws = %s feed = %g (%s) speed = %g (%s)", ws->get_its_id (), stix_get_measure_value (tech->get_feedrate ()), stix_get_unit_name (tech->get_feedrate ()),
			stix_get_measure_value (tech->get_spindle ()), stix_get_unit_name (tech->get_spindle ()));

    RoseDesign *des = ws->getRootObject()->design();

    Boring *op = Boring::newInstance (des);
    op->put_its_id (ws->get_its_id());						    // required but not usually shown
    op->put_its_tool (ROSE_CAST (stp_machining_tool, tool->getRootObject()));	    // Cast is annoying but necessary because tool is abstract
    op->put_its_machine_functions (funs->getRoot());
    op->put_its_technology (tech->getRoot ());
    if (cutting_depth > 0)
	op->put_cutting_depth (stix_make_measure_item (des, cutting_depth, length_unit));
    if (depth_of_testcut > 0)
	op->put_depth_of_testcut (stix_make_measure_item (des, depth_of_testcut, length_unit));
    if (previous_diameter > 0)
	op->put_previous_diameter (stix_make_measure_item (des, previous_diameter, length_unit));
    if (dwell_time_bottom > 0)
	op->put_dwell_time_bottom (stix_make_measure_item (des, dwell_time_bottom, time_unit));
    op->put_spindle_stop_at_bottom (spindle_stop_at_bottom);	    // EN_SPINDLE_NONSTOP, EN_SPINDLE_STOP_AT_BOTTOM
    StixVctr waiting = StixVctr (waiting_position_x, waiting_position_y, waiting_position_z);
    op->put_waiting_position(waiting.makePtIn (des, ""));
    if (feed_on_retract > 0)
	op->put_feed_on_retract (stix_make_ratio_measure_item (des, feed_on_retract, stixunit_ratio));

    ws->put_its_operation (op->getRoot());
    return op;
}

// Additional optional parameters for milling and drilling
Machining_operation_IF	*set_operation_overcut_length (Milling_type_operation_IF *op, double length, StixUnit length_unit)
{
    if (op == NULL || length_unit == NULL) {
	REPORT_ERROR ("Set operation overcut length: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Overcut length for operation %s set to %g (%s)", op->get_its_id (), length, stix_get_unit_name (length_unit));

    RoseDesign *des = op->getRootObject()->design();
    op->put_overcut_length (stix_make_measure_item (des, length, length_unit));
    return op;
}


// Approach and retract strategies for milling operations
Milling_type_operation_IF *set_plunge_toolaxis_approach (Milling_type_operation_IF *op)
{
    if (op == NULL) {
	REPORT_ERROR ("Set plunge tool axis approach: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Approach for operation %s set to %s", op->get_its_id (), "plunge tool axis");

    RoseDesign *des = op->getRootObject()->design();
    Plunge_toolaxis *approach = Plunge_toolaxis::newInstance (des);
    op->put_approach (approach->getRoot());
    return op;
}

Milling_type_operation_IF *set_plunge_ramp_approach (Milling_type_operation_IF *op, double angle, StixUnit degrees_or_radians)
{
    if (op == NULL) {
	REPORT_ERROR ("Set plunge ramp approach: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Approach for operation %s set to %s", op->get_its_id (), "plunge ramp");

    RoseDesign *des = op->getRootObject()->design();
    Plunge_ramp *approach = Plunge_ramp::newInstance (des);
    approach->put_angle (stix_make_angle_measure_item (des, angle, degrees_or_radians));
    op->put_approach (approach->getRoot());
    return op;
}

Milling_type_operation_IF *set_plunge_zigzag_approach (
	Milling_type_operation_IF *op, double angle, double width, StixUnit degrees_or_radians, StixUnit length_unit)
{
    if (op == NULL || length_unit == NULL || degrees_or_radians == NULL) {
	REPORT_ERROR ("Set plunge zigzag approach: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Approach for operation %s set to %s width = %g (%s), angle = %g (%s)", op->get_its_id (), "plunge zigzag",
			    width, stix_get_unit_name (length_unit), angle, stix_get_unit_name (degrees_or_radians));

    RoseDesign *des = op->getRootObject()->design();
    Plunge_zigzag *approach = Plunge_zigzag::newInstance (des);
    approach->put_angle (stix_make_angle_measure_item (des, angle, degrees_or_radians));
    approach->put_width (stix_make_measure_item (des, width, length_unit));
    op->put_approach (approach->getRoot());
    return op;
}

Milling_type_operation_IF *set_plunge_helix_approach (
	Milling_type_operation_IF *op, double angle, double radius, StixUnit degrees_or_radians, StixUnit length_unit)
{
    if (op == NULL || length_unit == NULL || degrees_or_radians == NULL) {
	REPORT_ERROR ("Set plunge helix approach: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Approach for operation %s set to %s radius = %g (%s), angle = %g (%s)", op->get_its_id (), "plunge helix",
			    radius, stix_get_unit_name (length_unit), angle, stix_get_unit_name (degrees_or_radians));

    RoseDesign *des = op->getRootObject()->design();
    Plunge_helix *approach = Plunge_helix::newInstance (des);
    approach->put_angle (stix_make_angle_measure_item (des, angle, degrees_or_radians));
    approach->put_radius (stix_make_measure_item (des, radius, length_unit));
    op->put_approach (approach->getRoot());
    return op;
}

Milling_type_operation_IF *set_air_angle_approach (
    Milling_type_operation_IF *op, double angle, double travel_length, StixUnit degrees_or_radians, StixUnit length_unit)
{
    if (op == NULL || length_unit == NULL || degrees_or_radians == NULL) {
	REPORT_ERROR ("Set air angle approach: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Approach for operation %s set to %s travel_length = %g (%s), angle = %g (%s)", op->get_its_id (), "air retract angle",
			    travel_length, stix_get_unit_name (length_unit), angle, stix_get_unit_name (degrees_or_radians));

    RoseDesign *des = op->getRootObject()->design();
    Ap_retract_angle *approach = Ap_retract_angle::newInstance (des);
    approach->put_angle (stix_make_angle_measure_item (des, angle, degrees_or_radians));
    approach->put_travel_length (stix_make_measure_item (des, travel_length, length_unit));
    op->put_approach (approach->getRoot());
    return op;
}

Milling_type_operation_IF *set_air_angle_retract (
	Milling_type_operation_IF *op, double angle, double travel_length, StixUnit degrees_or_radians, StixUnit length_unit)
{
    if (op == NULL || length_unit == NULL || degrees_or_radians == NULL) {
	REPORT_ERROR ("Set air angle retract: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Retract for operation %s set to %s travel_length = %g (%s), angle = %g (%s)", op->get_its_id (), "air angle",
			    travel_length, stix_get_unit_name (length_unit), angle, stix_get_unit_name (degrees_or_radians));

    RoseDesign *des = op->getRootObject()->design();
    Ap_retract_angle *retract = Ap_retract_angle::newInstance (des);
    retract->put_angle (stix_make_angle_measure_item (des, angle, degrees_or_radians));
    retract->put_travel_length (stix_make_measure_item (des, travel_length, length_unit));
    op->put_retract (retract->getRoot());
    return op;
}


Milling_type_operation_IF *set_air_tangent_approach (
    Milling_type_operation_IF *op, double radius, StixUnit length_unit)
{
    if (op == NULL || length_unit == NULL) {
	REPORT_ERROR ("Set air tangent approach: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Approach for operation %s set to %s radius = %g (%s)", op->get_its_id (), "air tangent",
			    radius, stix_get_unit_name (length_unit));

    RoseDesign *des = op->getRootObject()->design();
    Ap_retract_tangent *approach = Ap_retract_tangent::newInstance (des);
    approach->put_radius (stix_make_measure_item (des, radius, length_unit));
    op->put_approach (approach->getRoot());
    return op;
}

Milling_type_operation_IF *set_air_tangent_retract (
	Milling_type_operation_IF *op, double radius, StixUnit length_unit)
{
    if (op == NULL || length_unit == NULL) {
	REPORT_ERROR ("Set air angle retract: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Retract for operation %s set to %s radius = %g (%s)", op->get_its_id (), "air tangent",
			    radius, stix_get_unit_name (length_unit));

    RoseDesign *des = op->getRootObject()->design();
    Ap_retract_tangent *retract = Ap_retract_tangent::newInstance (des);
    retract->put_radius (stix_make_measure_item (des, radius, length_unit));
    op->put_retract (retract->getRoot());
    return op;
}

Milling_type_operation_IF *set_along_path_approach (Milling_type_operation_IF *op, Toolpath_IF *path)
{
    if (op == NULL || path == NULL) {
	REPORT_ERROR ("Set along path approach: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Approach for operation %s set to path %s", op->get_its_id (), path->get_its_id ());

    RoseDesign *des = op->getRootObject()->design();
    Along_path *approach = Along_path::newInstance (des);

    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    RoseObject *obj = NULL;
    if (center) {
	approach->add_path (center->getRoot ());
    }
    else if (contact) {
	approach->add_path (contact->getRoot ());
    }
    else {
	REPORT_ERROR ("Approach along path : path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return false;
    }

    op->put_approach (approach->getRoot());
    return op;
}

Milling_type_operation_IF *add_path_to_approach (Milling_type_operation_IF *op, Toolpath_IF *path)
{
    if (op == NULL || path == NULL) {
	REPORT_ERROR ("Add path to approach: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Appraoch for operation %s adding path %s", op->get_its_id (), path->get_its_id ());

    Along_path *approach = Along_path::find(op->get_approach());
    if (approach == NULL) {
	REPORT_ERROR ("Add path to approach: approach has no previous path");
	return NULL;
    }

    RoseDesign *des = op->getRootObject()->design();

    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    RoseObject *obj = NULL;
    if (center) {
	approach->add_path (center->getRoot());
    }
    else if (contact) {
	approach->add_path (contact->getRoot ());
    }
    else {
	REPORT_ERROR ("Approach along path : path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return false;
    }

     return op;
}

Milling_type_operation_IF *set_along_path_retract (Milling_type_operation_IF *op, Toolpath_IF *path)
{
    if (op == NULL || path == NULL) {
	REPORT_ERROR ("Set along path retract: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Retract for operation %s set to path %s", op->get_its_id (), path->get_its_id ());

    RoseDesign *des = op->getRootObject()->design();
    Along_path *retract = Along_path::newInstance (des);

    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    RoseObject *obj = NULL;
    if (center) {
	retract->add_path (center->getRoot());
    }
    else if (contact) {
	retract->add_path (contact->getRoot ());
    }
    else {
	REPORT_ERROR ("Retract along path : path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return false;
    }

    op->put_retract (retract->getRoot());
    return op;
}

Milling_type_operation_IF *add_path_to_retract (Milling_type_operation_IF *op, Toolpath_IF *path)
{
    if (op == NULL || path == NULL) {
	REPORT_ERROR ("Add path to retract: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Retract for operation %s adding path %s", op->get_its_id (), path->get_its_id ());

    Along_path *retract = Along_path::find(op->get_retract());
    if (retract == NULL) {
	REPORT_ERROR ("Add path to retract: retract has no previous path");
	return NULL;
    }

    RoseDesign *des = op->getRootObject()->design();

    Cutter_location_trajectory *center = Cutter_location_trajectory::find (path->getRootObject ());
    Cutter_contact_trajectory *contact = Cutter_contact_trajectory::find (path->getRootObject ());

    RoseObject *obj = NULL;
    if (center) {
	retract->add_path (center->getRoot());
    }
    else if (contact) {
	retract->add_path (contact->getRoot ());
    }
    else {
	REPORT_ERROR ("Retract along path : path is not Cutter_location_trajectory or Cutter_center_trajectory");
	return false;
    }

     return op;
}

// Optional description for the tool orientation at the start and end of each operation for five axis machining
Milling_type_operation_IF *set_operation_start_orientation (Milling_type_operation_IF *op, double i, double j, double k)
{
    if (op == NULL) {
	REPORT_ERROR ("Set start orientation: function called with NULL arguments");
	return NULL;
    }
    if (Approach_retract_strategy_IF::find (op->get_approach()) == NULL) {
	REPORT_ERROR ("Set start orientation: strategy must be defined before tool orientation");
	return NULL;
    }
    TRACE_CALL ("Set start orientation for operation %s to  (%g, %g, %g)", op->get_its_id (), i, j, k);

    RoseDesign *des = op->getRootObject()->design();
    Approach_retract_strategy_IF * approach = Approach_retract_strategy_IF::find (op->get_approach());
    StixVctr dir = StixVctr (i, j, k);
    approach->put_tool_orientation (dir.makeDirIn (des, ""));
    return op;
}

Milling_type_operation_IF *set_operation_end_orientation (Milling_type_operation_IF *op, double i, double j, double k)
{
    if (op == NULL) {
	REPORT_ERROR ("Set end orientation: function called with NULL arguments");
	return NULL;
    }
    if (Approach_retract_strategy_IF::find (op->get_retract()) == NULL) {
	REPORT_ERROR ("Set end orientation: strategy must be defined before tool orientation");
	return NULL;
    }
    TRACE_CALL ("Set end orientation for operation %s to (%g, %g, %g)", op->get_its_id (), i, j, k);

    RoseDesign *des = op->getRootObject()->design();
    Approach_retract_strategy_IF * retract = Approach_retract_strategy_IF::find (op->get_retract());
    StixVctr dir = StixVctr (i, j, k);
    retract->put_tool_orientation (dir.makeDirIn (des, ""));
    return op;
}

Machining_operation_IF	*set_operation_start_point (Machining_operation_IF *op, double x, double y, double z)
{
    if (op == NULL) {
	REPORT_ERROR ("Set operation start point: function called with NULL arguments");
	return NULL;
    }
    RoseDesign *des = op->getRootObject()->design();
    if (get_geometric_context (des) == NULL) {
	REPORT_ERROR ("Set operation start point: no units set (see set_toolpath_units)");
	return NULL;
    }
    TRACE_CALL ("Set start point for operation %s to  (%g, %g, %g) unit = %s", op->get_its_id (), x, y, z, stix_get_unit_name (stix_get_context_length_unit (get_geometric_context (des))));

    StixVctr dir = StixVctr (x, y, z);
    op->put_start_point (dir.makePtIn (des, ""));
    return op;
}

Machining_operation_IF	*set_operation_retract_distance (Machining_operation_IF *op, double retract_distance, StixUnit length_unit)
{
    if (op == NULL || length_unit == NULL) {
	REPORT_ERROR ("Set operation retract distance: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Set retract distance for operation %s to  %g", op->get_its_id (), retract_distance);

    RoseDesign *des = op->getRootObject()->design();
    op->put_retract_plane (stix_make_measure_item (des, retract_distance, length_unit));
    return op;
}

Two5d_milling_operation_IF *set_center_strategy (Two5d_milling_operation_IF * op)
{
    if (op == NULL) {
	REPORT_ERROR ("Set center strategy: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Strategy for operation %s set to %s", op->get_its_id (), "center milling");

    RoseDesign *des = op->getRootObject()->design();
    Center_milling *strategy = Center_milling::newInstance (des);
    op->put_its_machining_strategy (strategy->getRoot());
    return op;
}

Two5d_milling_operation_IF *set_explicit_strategy (Two5d_milling_operation_IF * op)
{
    if (op == NULL) {
	REPORT_ERROR ("Set explicit strategy: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Strategy for operation %s set to %s", op->get_its_id (), "explicit milling");

    RoseDesign *des = op->getRootObject()->design();
    Explicit_strategy *strategy = Explicit_strategy::newInstance (des);
    op->put_its_machining_strategy (strategy->getRoot());
    return op;
}


Two5d_milling_operation_IF *set_contour_spiral_strategy (
	Two5d_milling_operation_IF * op, const char *cutmode, const char *rot_direction, double overlap_ratio)
{
    if (op == NULL || cutmode == NULL || rot_direction == NULL) {
	REPORT_ERROR ("Set contour spiral strategy: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Strategy for operation %s set to %s cutmode = %s, rotation direction = %s", op->get_its_id (), "contour spiral milling",
		    cutmode, rot_direction);

    RoseDesign *des = op->getRootObject()->design();
    Contour_spiral *strategy = Contour_spiral::newInstance (des);
    strategy->put_cutmode (cutmode);				    // EN_CLIMB, EN_CONVENTIONAL
    strategy->put_rotation_direction (rot_direction);		    // EN_CLOCKWISE, EN_COUNTERCLOCKWISE
    strategy->put_overlap (stix_make_ratio_measure_item (des, overlap_ratio, stixunit_ratio));
    op->put_its_machining_strategy (strategy->getRoot());
    return op;
}


Two5d_milling_operation_IF *set_contour_parallel_strategy (
	Two5d_milling_operation_IF * op, const char *cutmode, const char *rot_direction, double overlap_ratio)
{
    if (op == NULL || cutmode == NULL || rot_direction == NULL) {
	REPORT_ERROR ("Set contour parallel strategy: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Strategy for operation %s set to %s cutmode = %s, rotation direction = %s", op->get_its_id (), "contour parallel milling",
		    cutmode, rot_direction);

    RoseDesign *des = op->getRootObject()->design();
    Contour_parallel *strategy = Contour_parallel::newInstance (des);
    strategy->put_cutmode (cutmode);				    // EN_CLIMB, EN_CONVENTIONAL
    strategy->put_rotation_direction (rot_direction);		    // EN_CLOCKWISE, EN_COUNTERCLOCKWISE
    strategy->put_overlap (stix_make_ratio_measure_item (des, overlap_ratio, stixunit_ratio));
    op->put_its_machining_strategy (strategy->getRoot());
    return op;
}

Two5d_milling_operation_IF *set_contour_bidirectional_strategy (
	Two5d_milling_operation_IF * op, const char *spiral_cutmode, 
	double i, double j, double k, const char *rot_direction, const char * stepover_direction, double overlap_ratio)
{
    if (op == NULL || spiral_cutmode == NULL || rot_direction == NULL || stepover_direction == NULL) {
	REPORT_ERROR ("Set contour bidirectional strategy: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Strategy for operation %s set to %s spiral_cutmode = %s, rotation direction = %s stepover_direction = %s", 
	op->get_its_id (), "contour bidirectional milling",
		    spiral_cutmode, rot_direction, stepover_direction);

    RoseDesign *des = op->getRootObject()->design();
    Contour_bidirectional *strategy = Contour_bidirectional::newInstance (des);
    strategy->put_spiral_cutmode (spiral_cutmode);		    // EN_CLIMB, EN_CONVENTIONAL
    strategy->put_rotation_direction (rot_direction);		    // EN_CLOCKWISE, EN_COUNTERCLOCKWISE
    strategy->put_stepover_direction (stepover_direction);	    // EN_LEFT, EN_RIGHT
    StixVctr feed_direction = StixVctr (i, j, k);
    strategy->put_feed_direction (feed_direction.makeDirIn (des, ""));
    strategy->put_overlap (stix_make_ratio_measure_item (des, overlap_ratio, stixunit_ratio));
    op->put_its_machining_strategy (strategy->getRoot());
    return op;
}
    
Two5d_milling_operation_IF *set_bidirectional_strategy (Two5d_milling_operation_IF * op,  
	double i, double j, double k, const char * stepover_direction, const char *connection_strategy, double overlap_ratio)
{
    if (op == NULL || stepover_direction == NULL || connection_strategy == NULL) {
	REPORT_ERROR ("Set bidirectional strategy: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Strategy for operation %s set to %s stepover direction = %s, stroke connection strategy = %s", 
	op->get_its_id (), "bidirectional milling",
		    stepover_direction, connection_strategy);

    RoseDesign *des = op->getRootObject()->design();
    Bidirectional *strategy = Bidirectional::newInstance (des);
    strategy->put_stepover_direction (stepover_direction);		    // EN_LEFT, EN_RIGHT
    strategy->put_its_stroke_connection_strategy (connection_strategy);	    // EN_STRAIGHT_LINE, EN_LEFT_SHIFT_PLUNGE, EN_DEGOUGE, EN_LOOP_BACK
    StixVctr feed_direction = StixVctr (i, j, k);
    strategy->put_feed_direction (feed_direction.makeDirIn (des, ""));
    strategy->put_overlap (stix_make_ratio_measure_item (des, overlap_ratio, stixunit_ratio));
    op->put_its_machining_strategy (strategy->getRoot());
    return op;
}

Two5d_milling_operation_IF *set_bidirectional_contour_strategy (
	Two5d_milling_operation_IF * op,  const char *spiral_cutmode,  
	double i, double j, double k, const char *rot_direction, const char * stepover_direction, double overlap_ratio)
{
    if (op == NULL || spiral_cutmode == NULL || rot_direction == NULL || stepover_direction == NULL) {
	REPORT_ERROR ("Set bidirectional contour strategy: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Strategy for operation %s set to %s spiral_cutmode = %s, rotation direction = %s stepover_direction = %s", 
	op->get_its_id (), "bidirectional contour milling",
		    spiral_cutmode, rot_direction, stepover_direction);

    RoseDesign *des = op->getRootObject()->design();
    Bidirectional_contour *strategy = Bidirectional_contour::newInstance (des);
    strategy->put_spiral_cutmode (spiral_cutmode);		    // EN_CLIMB, EN_CONVENTIONAL
    strategy->put_rotation_direction (rot_direction);		    // EN_CLOCKWISE, EN_COUNTERCLOCKWISE
    strategy->put_stepover_direction (stepover_direction);	    // EN_LEFT, EN_RIGHT
    StixVctr feed_direction = StixVctr (i, j, k);
    strategy->put_feed_direction (feed_direction.makeDirIn (des, ""));
    strategy->put_overlap (stix_make_ratio_measure_item (des, overlap_ratio, stixunit_ratio));
    op->put_its_machining_strategy (strategy->getRoot());
    return op;
}

Two5d_milling_operation_IF *set_unidirectional_strategy (
	Two5d_milling_operation_IF * op,  const char *cutmode, double i, double j, double k, double overlap_ratio)
{
    if (op == NULL || cutmode == NULL) {
	REPORT_ERROR ("Set unidirectional strategy: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Strategy for operation %s set to %s cutmode = %s", op->get_its_id (), "unidirectional milling",
		    cutmode);

    RoseDesign *des = op->getRootObject()->design();
    Unidirectional *strategy = Unidirectional::newInstance (des);
    strategy->put_cutmode (cutmode);		    // EN_CLIMB, EN_CONVENTIONAL
    StixVctr feed_direction = StixVctr (i, j, k);
    strategy->put_feed_direction (feed_direction.makeDirIn (des, ""));
    strategy->put_overlap (stix_make_ratio_measure_item (des, overlap_ratio, stixunit_ratio));
    op->put_its_machining_strategy (strategy->getRoot());
    return op;
}

Two5d_milling_operation_IF *set_strategy_disallow_multiple_passes (Two5d_milling_operation_IF * op)
{
    if (op == NULL) {
	REPORT_ERROR ("Set strategy disallow multiple passes: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Strategy for operation %s set to disallow multiple passes = %s", op->get_its_id ());

    Two5d_milling_strategy_IF * strategy = Two5d_milling_strategy_IF::find (op->get_its_machining_strategy ());		    
    if (strategy == NULL) {
	REPORT_ERROR ("Set strategy disallow multiple passes: strategy must be set before you can disallow multiple passes");
	return NULL;
    }

    // EN_MULTIPLE_PASSES_ALLOWED (default no need to set), EN_MULTIPLE_PASSES_NOT_ALLOWED
    strategy->put_allow_multiple_passes (EN_MULTIPLE_PASSES_NOT_ALLOWED);
    return op;
}

// Functions to add CC2 structural value to the data
Machining_workingstep *set_workingstep_process_feature (Machining_workingstep *ws, Manufacturing_feature_IF *feature)
{
    if (ws == NULL || feature == NULL) {
	REPORT_ERROR ("Set workingstep process feature: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Seting process feature for workingstep %s to feature %s", ws->get_its_id (), feature->get_its_id ());

    ws->put_its_feature (ROSE_CAST (stp_shape_aspect, feature->getRootObject()));
    return ws;
}

Machining_workingstep *add_workingstep_final_feature (Machining_workingstep *ws, Manufacturing_feature_IF *feature)
{
    if (ws == NULL || feature == NULL) {
	REPORT_ERROR ("Add workingstep final feature: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Adding final feature %s to workingstep %s", feature->get_its_id (), ws->get_its_id ());

    ws->add_final_features (ROSE_CAST (stp_shape_aspect, feature->getRootObject()));
    return ws;
}

Executable_IF *set_executable_as_is_workpiece (Executable_IF *exe, Workpiece *piece) 
{
    if (exe == NULL || piece == NULL) {
	REPORT_ERROR ("Set executable as_is workpiece: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Adding as_is workpiece %s to executable %s", piece->get_its_id (), exe->get_its_id ());

    exe->put_as_is_geometry (piece->getRoot ());
    return exe;
}

Executable_IF *set_executable_to_be_workpiece (Executable_IF *exe, Workpiece *piece) 
{
    if (exe == NULL || piece == NULL) {
	REPORT_ERROR ("Set executable to_be workpiece: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Adding to_be workpiece %s to executable %s", piece->get_its_id (), exe->get_its_id ());

    exe->put_to_be_geometry (piece->getRoot ());
    return exe;
}

Executable_IF *set_executable_removal_workpiece (Executable_IF *exe, Workpiece *piece) 
{
    if (exe == NULL || piece == NULL) {
	REPORT_ERROR ("Set executable removal workpiece: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Adding removal workpiece %s to executable %s", piece->get_its_id (), exe->get_its_id ());

    exe->put_removal_geometry (piece->getRoot ());
    return exe;
}

Machining_workingstep *set_workingstep_toolpath_orientation (Machining_workingstep *ws, StixMtrx M) 
{
    if (ws == NULL) {
	REPORT_ERROR ("Set workingstep toolpath orientation: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Setting toolpath orientation workpingstep %s to (%g, %g, %g) z=(%g, %g, %g) x=(%g, %g, %g)", ws->get_its_id (), M.w().get(0), M.w().get(1), M.w().get(2), M.z().get(0), M.z().get(1), M.z().get(2), M.x().get(0), M.x().get(1), M.x().get(2));

    RoseDesign *des = ws->getRootObject()->design();

    stp_axis2_placement_3d *orientation = M.makeAp3dIn (des);

    ws->put_toolpath_orientation (orientation);
    return ws;
}

Machining_workingstep *set_workingstep_security_plane (Machining_workingstep *ws, StixMtrx M)
{
    if (ws == NULL) {
	REPORT_ERROR ("Set workingstep security plane: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Setting security plane of workingstep %s to (%g, %g, %g) z=(%g, %g, %g) x=(%g, %g, %g)", ws->get_its_id (), M.w().get(0), M.w().get(1), M.w().get(2), M.z().get(0), M.z().get(1), M.z().get(2), M.x().get(0), M.x().get(1), M.x().get(2));

    RoseDesign *des = ws->getRootObject()->design();

    stp_axis2_placement_3d *orientation = M.makeAp3dIn (des);
    stp_plane *plane = pnewIn (des) stp_plane;
    plane->position (orientation);
    ws->put_its_secplane (plane);
    return ws;
}

// Make a new workplan
Workplan *new_workplan (Workplan *parent_plan, const char *plan_name)
{
    if (parent_plan == NULL || plan_name == NULL) {
	REPORT_ERROR ("new workplan: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Workplan name = %s being added as sub-plan of workplan name = %s", plan_name, parent_plan->get_its_id());
 
    RoseDesign *des = parent_plan->getRootObject()->design();
    Workplan *plan = Workplan::newInstance (des);
    plan->put_its_id (plan_name);
    parent_plan->add_its_elements (plan->getRoot());

    return plan;
}

Workplan *new_workplan (Selective *parent_selective, const char *plan_name)
{
    if (parent_selective == NULL || plan_name == NULL) {
	REPORT_ERROR ("new workplan: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Workplan name = %s being added as sub-plane of selective name = %s", plan_name, parent_selective->get_its_id());
 
    RoseDesign *des = parent_selective->getRootObject()->design();
    Workplan *plan = Workplan::newInstance (des);
    plan->put_its_id (plan_name);
    parent_selective->add_its_elements (plan->getRoot());

    return plan;
}

// Make a new selective
Selective *new_selective (Workplan *parent_plan, const char *selective_name)
{
    if (parent_plan == NULL || selective_name == NULL) {
	REPORT_ERROR ("new selective: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Selective name = %s being added as sub-plan of workplan name = %s", selective_name, parent_plan->get_its_id());
 
    RoseDesign *des = parent_plan->getRootObject()->design();
    Selective *plan = Selective::newInstance (des);
    plan->put_its_id (selective_name);
    parent_plan->add_its_elements (plan->getRoot());

    return plan;
}

Selective *new_selective (Selective *parent_selective, const char *selective_name)
{
    if (parent_selective == NULL || selective_name == NULL) {
	REPORT_ERROR ("new selective: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Selective name = %s being added as sub-plane of selective name = %s", selective_name, parent_selective->get_its_id());
 
    RoseDesign *des = parent_selective->getRootObject()->design();
    Selective *plan = Selective::newInstance (des);
    plan->put_its_id (selective_name);
    parent_selective->add_its_elements (plan->getRoot());

    return plan;
}

// Insert and remove executabes into workplans and selectives
Workplan *insert_executable_after (Workplan *plan, Executable_IF *current, Executable_IF *insert)
{
    if (plan == NULL || insert == NULL || current == NULL) {
	REPORT_ERROR ("insert exectuable after: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Executable name = %s being inserted into workplan name = %s after executable %s", insert->get_its_id(), plan->get_its_id(), current->get_its_id());
  
    Workplan::Its_elements_IFCursor cur(plan);
    Workplan::Its_elements_IF * el;
    RoseObject *obj = current->getRootObject();
    stp_machining_process_executable *ex = ROSE_CAST (stp_machining_process_executable, insert->getRootObject ());

    cur.rewind();
    while (el = cur.next()) {
	if (obj == el->getValue()) {
	    double index, new_index;
	    index = el->getIndex ();
	    el = cur.next();
	    if (el)
		new_index = (index + el->getIndex ()) / 2;
	    else
		new_index = index + 1;
	    plan->add_its_elements (ex, new_index);	
	    ex = NULL;
	    break;
	}
    }

    if (ex != NULL)
	plan->add_its_elements (ex);
    else
	plan->its_elements.sort();

    return plan;
}

Workplan *remove_executable (Workplan *plan, Executable_IF *remove)
{
    if (plan == NULL || remove == NULL) {
	REPORT_ERROR ("remove executable: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Executable name = %s being removed from workplan name = %s", remove->get_its_id(), plan->get_its_id());

    Workplan::Its_elements_IFCursor cur(plan);
    Workplan::Its_elements_IF * el;
    RoseObject *obj = remove->getRootObject();

    cur.rewind();
    while (el = cur.next()) {
	if (obj == el->getValue()) {
	    plan->remove_its_elements (el);	// changes cursor best to start again
	    cur.rewind ();
	}
    }

    plan->its_elements.sort ();
    return plan;
}

Selective *insert_executable_after (Selective *plan, Executable_IF *current, Executable_IF *insert)
{
    if (plan == NULL || insert == NULL || current == NULL) {
	REPORT_ERROR ("insert exectuable after: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Executable name = %s being inserted into selective name = %s after executable %s", insert->get_its_id(), plan->get_its_id(), current->get_its_id());
 
    Selective::Its_elements_IFCursor cur(plan);
    Selective::Its_elements_IF * el;
    RoseObject *obj = current->getRootObject();
    stp_machining_process_executable *ex = ROSE_CAST (stp_machining_process_executable, insert->getRootObject ());

    cur.rewind();
    while (el = cur.next()) {
	if (obj == el->getValue()) {
	    double index, new_index;
	    index = el->getIndex ();
	    el = cur.next();
	    if (el)
		new_index = (index + el->getIndex ()) / 2;
	    else
		new_index = index + 1;
	    plan->add_its_elements (ex, new_index);	
	    ex = NULL;
	    break;
	}
    }

    if (ex != NULL)
	plan->add_its_elements (ex);
    else
	plan->its_elements.sort();

    return plan;
}

Selective *remove_executable (Selective *plan, Executable_IF *remove)
{
    if (plan == NULL || remove == NULL) {
	REPORT_ERROR ("remove executable: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Executable name = %s being removed from selective name = %s", remove->get_its_id(), plan->get_its_id());

    Selective::Its_elements_IFCursor cur(plan);
    Selective::Its_elements_IF * el;
    RoseObject *obj = remove->getRootObject();

    cur.rewind();
    while (el = cur.next()) {
	if (obj == el->getValue()) {
	    plan->remove_its_elements (el);	// changes cursor best to start again
	    cur.rewind ();
	}
    }

    plan->its_elements.sort ();
    return plan;
}

// Make a workingstep (three flavors)
Machining_workingstep *new_workingstep (Selective *parent_selective, const char *ws_name)
{
    if (parent_selective == NULL || ws_name == NULL) {
	REPORT_ERROR ("New workingstep: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Workingstep name = %s being added to selective name = %s", ws_name, parent_selective->get_its_id());
    
    RoseDesign *des = parent_selective->getRootObject()->design();
    Machining_workingstep *ws = Machining_workingstep::newInstance (des);
    ws->put_its_id (ws_name);

    Manufacturing_feature_IF *default = find_default_feature_on_default_workpiece (des);
    ws->put_its_feature (ROSE_CAST (stp_shape_aspect, default->getRootObject()));
    parent_selective->add_its_elements (ws->getRoot());

    return ws;
}

// Orphan until inserted -- strict garbage collection will remove when file written
Machining_workingstep *orphan_workingstep (RoseDesign *des, const char *ws_name)						
{
    if (des == NULL || ws_name == NULL) {
	REPORT_ERROR ("New workingstep: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Workingstep name = %s being added as orphan to design name = %s", ws_name, des->name());
    
    Machining_workingstep *ws = Machining_workingstep::newInstance (des);
    ws->put_its_id (ws_name);

    Manufacturing_feature_IF *default = find_default_feature_on_default_workpiece (des);
    ws->put_its_feature (ROSE_CAST (stp_shape_aspect, default->getRootObject()));

    return ws;
}

Workplan *orphan_workplan (RoseDesign *des, const char *plan_name)
{
    if (des == NULL || plan_name == NULL) {
	REPORT_ERROR ("New workplan: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Workplan name = %s being added as orphan to design name = %s", plan_name, des->name());
    
    Workplan *plan = Workplan::newInstance (des);
    plan->put_its_id (plan_name);

    return plan;
}
	
Selective *oprhan_selective (RoseDesign *des, const char *plan_name)
{
    if (des == NULL || plan_name == NULL) {
	REPORT_ERROR ("New workplan: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Workplan name = %s being added as orphan to design name = %s", plan_name, des->name());
    
    Selective *plan = Selective::newInstance (des);
    plan->put_its_id (plan_name);

    return plan;
}

Executable_IF *enable_executable (Executable_IF *ws_or_plan_or_selective)
{
    if (ws_or_plan_or_selective == NULL) {
	REPORT_ERROR ("Enable executable: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Executable name = %s being enabled = %s", ws_or_plan_or_selective->get_its_id());
    
    ws_or_plan_or_selective->put_enabled ("enabled");
    return ws_or_plan_or_selective;
}

Executable_IF *disable_executable (Executable_IF *ws_or_plan_or_selective)
{
    if (ws_or_plan_or_selective == NULL) {
	REPORT_ERROR ("Disable executable: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Executable name = %s being disabled = %s", ws_or_plan_or_selective->get_its_id());
    
    ws_or_plan_or_selective->put_enabled ("disabled");
    return ws_or_plan_or_selective;
}

// Define a Setup for a workplan, define a fixture for a setup and define placements for the workpiece in the fixture and the fixture on the machine
Setup *workplan_setup (Workplan *plan, StixMtrx M)
{
    if (plan == NULL) {
	REPORT_ERROR ("workplan setup: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Setting setup of workplan %s to (%g, %g, %g) z=(%g, %g, %g) x=(%g, %g, %g)", plan->get_its_id (), M.w().get(0), M.w().get(1), M.w().get(2), M.z().get(0), M.z().get(1), M.z().get(2), M.x().get(0), M.x().get(1), M.x().get(2));

    RoseDesign *des = plan->getRootObject()->design();

    Setup *up = Setup::find (plan->get_its_setup ());
    if (up == NULL) {
	up = Setup::newInstance (des);
	plan->put_its_setup (up->getRoot ());
	up->put_its_id (plan->get_its_id());
    }

    up->put_its_origin (M.makeAp3dIn (des));

    return up;
}

Setup *fixture_geometry (Setup *setup, Workpiece *model)
{
    if (setup == NULL || model == NULL) {
	REPORT_ERROR ("fixture geometry: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Setting fixture for setup %s to model %s", setup->get_its_id(), model->get_its_id());

    RoseDesign *des = setup->getRootObject()->design();

    Fixture_usage * fu = Fixture_usage::find (setup->get_its_fixture_usage());
    if (fu == NULL) {
	fu = Fixture_usage::newInstance (des);
	fu->put_its_id (setup->get_its_id());
	setup->put_its_fixture_usage (fu->getRoot ());
    }

    fu->put_its_product (model->getRoot ());
    return setup;
}

Setup *fixture_mount_on_machine (Setup *setup,  StixMtrx M)
{
    if (setup == NULL) {
	REPORT_ERROR ("fixture mount on machine: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Setting fixture on machine for setup %s to (%g, %g, %g) z=(%g, %g, %g) x=(%g, %g, %g)", setup->get_its_id (), M.w().get(0), M.w().get(1), M.w().get(2), M.z().get(0), M.z().get(1), M.z().get(2), M.x().get(0), M.x().get(1), M.x().get(2));

    RoseDesign *des = setup->getRootObject()->design();

    Fixture_usage * fu = Fixture_usage::find (setup->get_its_fixture_usage());
    if (fu == NULL) {
	fu = Fixture_usage::newInstance (des);
	fu->put_its_id (setup->get_its_id());
	setup->put_its_fixture_usage (fu->getRoot ());
    }

    fu->put_mount_reference (M.makeAp3dIn (des));

    return setup;
}

Setup *workpiece_mount_on_fixture (Setup *setup,  StixMtrx M)
{
    if (setup == NULL) {
	REPORT_ERROR ("workpiece mount on fixture: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Setting workpiece mount point on fixture for setup %s to (%g, %g, %g) z=(%g, %g, %g) x=(%g, %g, %g)", setup->get_its_id (), M.w().get(0), M.w().get(1), M.w().get(2), M.z().get(0), M.z().get(1), M.z().get(2), M.x().get(0), M.x().get(1), M.x().get(2));

    RoseDesign *des = setup->getRootObject()->design();

    Fixture_usage * fu = Fixture_usage::find (setup->get_its_fixture_usage());
    if (fu == NULL) {
	fu = Fixture_usage::newInstance (des);
	fu->put_its_id (setup->get_its_id());
	setup->put_its_fixture_usage (fu->getRoot ());
    }

    fu->put_workpiece_reference (M.makeAp3dIn (des));

    return setup;
}

Setup *setup_security_plane (Setup* setup, StixMtrx M)
{
    if (setup == NULL) {
	REPORT_ERROR ("setup security plane: function called with NULL arguments");
	return NULL;
    }
    TRACE_CALL ("Setting security plane of setup %s to (%g, %g, %g) z=(%g, %g, %g) x=(%g, %g, %g)", setup->get_its_id (), M.w().get(0), M.w().get(1), M.w().get(2), M.z().get(0), M.z().get(1), M.z().get(2), M.x().get(0), M.x().get(1), M.x().get(2));

    RoseDesign *des = setup->getRootObject()->design();

    stp_plane *plane = pnewIn (des) stp_plane;
    plane->position (M.makeAp3dIn (des));
    setup->put_its_secplane (plane);

    return setup;
}
